@@ -36,12 +36,6 @@ void
 cli_help(int argc, char **argv)
 {
 	CommandLine command = root;
-
-	if (env_exists(PGCOPYDB_DEBUG))
-	{
-		command = root_with_debug;
-	}
-
 	(void) commandline_print_command_tree(&command, stdout);
 }
 
@@ -57,19 +51,13 @@ cli_print_version_getopts(int argc, char **argv)
 
 	static struct option long_options[] = {
 		{ "json", no_argument, NULL, 'J' },
-		{ "version", no_argument, NULL, 'V' },
-		{ "verbose", no_argument, NULL, 'v' },
-		{ "notice", no_argument, NULL, 'v' },
-		{ "debug", no_argument, NULL, 'd' },
-		{ "trace", no_argument, NULL, 'z' },
-		{ "quiet", no_argument, NULL, 'q' },
 		{ "help", no_argument, NULL, 'h' },
 		{ NULL, 0, NULL, 0 }
 	};
 	optind = 0;
 
 	/*
-	 * The only command lines that are using keeper_cli_getopt_pgdata are
+	 * The only command lines that are using cli_print_version_getopts are
 	 * terminal ones: they don't accept subcommands. In that case our option
 	 * parsing can happen in any order and we don't need getopt_long to behave
 	 * in a POSIXLY_CORRECT way.
@@ -78,7 +66,7 @@ cli_print_version_getopts(int argc, char **argv)
 	 */
 	unsetenv("POSIXLY_CORRECT");
 
-	while ((c = getopt_long(argc, argv, "JVvdzqh",
+	while ((c = getopt_long(argc, argv, "Jh",
 							long_options, &option_index)) != -1)
 	{
 		switch (c)
@@ -97,6 +85,7 @@ cli_print_version_getopts(int argc, char **argv)
 				break;
 			}
 
+			case '?':
 			default:
 			{
 				/*
@@ -121,13 +110,13 @@ cli_print_version(int argc, char **argv)
 	if (outputJSON)
 	{
 		JSON_Value *js = json_value_init_object();
-		JSON_Object *root = json_value_get_object(js);
+		JSON_Object *jsObj = json_value_get_object(js);
 
-		json_object_set_string(root, "pgcopydb", VERSION_STRING);
-		json_object_set_string(root, "pg_major", PG_MAJORVERSION);
-		json_object_set_string(root, "pg_version", PG_VERSION);
-		json_object_set_string(root, "pg_version_str", PG_VERSION_STR);
-		json_object_set_number(root, "pg_version_num", (double) PG_VERSION_NUM);
+		json_object_set_string(jsObj, "pgcopydb", VERSION_STRING);
+		json_object_set_string(jsObj, "pg_major", PG_MAJORVERSION);
+		json_object_set_string(jsObj, "pg_version", PG_VERSION);
+		json_object_set_string(jsObj, "pg_version_str", PG_VERSION_STR);
+		json_object_set_number(jsObj, "pg_version_num", (double) PG_VERSION_NUM);
 
 		(void) cli_pprint_json(js);
 	}
@@ -284,6 +273,28 @@ cli_copydb_getenv(CopyDBOptions *options)
 		}
 	}
 
+	if (env_exists(PGCOPYDB_RESTORE_JOBS))
+	{
+		char jobs[BUFSIZE] = { 0 };
+
+		if (get_env_copy(PGCOPYDB_RESTORE_JOBS, jobs, sizeof(jobs)))
+		{
+			if (!stringToInt(jobs, &options->restoreOptions.jobs) ||
+				options->restoreOptions.jobs < 1 ||
+				options->restoreOptions.jobs > 128)
+			{
+				log_fatal("Failed to parse PGCOPYDB_RESTORE_JOBS: \"%s\"",
+						  jobs);
+				++errors;
+			}
+		}
+		else
+		{
+			/* errors have already been logged */
+			++errors;
+		}
+	}
+
 	if (env_exists(PGCOPYDB_LARGE_OBJECTS_JOBS))
 	{
 		char jobs[BUFSIZE] = { 0 };
@@ -346,6 +357,29 @@ cli_copydb_getenv(CopyDBOptions *options)
 		}
 	}
 
+	/* check if --wal2json-numeric-as-string has been used */
+	if (env_exists(PGCOPYDB_WAL2JSON_NUMERIC_AS_STRING))
+	{
+		char wal2jsonNumericAsString[BUFSIZE] = { 0 };
+
+		if (!get_env_copy(PGCOPYDB_WAL2JSON_NUMERIC_AS_STRING,
+						  wal2jsonNumericAsString,
+						  sizeof(wal2jsonNumericAsString)))
+		{
+			/* errors have already been logged */
+			++errors;
+		}
+		else if (!parse_bool(wal2jsonNumericAsString,
+							 &(options->slot.wal2jsonNumericAsString)))
+		{
+			log_error("Failed to parse environment variable \"%s\" "
+					  "value \"%s\", expected a boolean (on/off)",
+					  PGCOPYDB_WAL2JSON_NUMERIC_AS_STRING,
+					  wal2jsonNumericAsString);
+			++errors;
+		}
+	}
+
 	/* when --drop-if-exists has not been used, check PGCOPYDB_DROP_IF_EXISTS */
 	if (!options->restoreOptions.dropIfExists)
 	{
@@ -395,7 +429,7 @@ cli_copydb_getenv(CopyDBOptions *options)
 		}
 	}
 
-	/* when --fail-fast has not been used, check PGCOPYDB_FAIL_FAST */
+	/* when --skip-vacuum has not been used, check PGCOPYDB_SKIP_VACUUM */
 	if (!options->skipVacuum)
 	{
 		if (env_exists(PGCOPYDB_SKIP_VACUUM))
@@ -713,6 +747,7 @@ cli_copy_db_getopts(int argc, char **argv)
 		{ "no-role-passwords", no_argument, NULL, 'P' },
 		{ "no-owner", no_argument, NULL, 'O' },       /* pg_restore -O */
 		{ "no-comments", no_argument, NULL, 'X' },
+		{ "restore-jobs", required_argument, NULL, 'j' },      /* pg_restore --jobs */
 		{ "no-acl", no_argument, NULL, 'x' }, /* pg_restore -x */
 		{ "skip-blobs", no_argument, NULL, 'B' },
 		{ "skip-large-objects", no_argument, NULL, 'B' },
@@ -731,6 +766,7 @@ cli_copy_db_getopts(int argc, char **argv)
 		{ "snapshot", required_argument, NULL, 'N' },
 		{ "follow", no_argument, NULL, 'f' },
 		{ "plugin", required_argument, NULL, 'p' },
+		{ "wal2json-numeric-as-string", no_argument, NULL, 'w' },
 		{ "slot-name", required_argument, NULL, 's' },
 		{ "origin", required_argument, NULL, 'o' },
 		{ "create-slot", no_argument, NULL, 't' },
@@ -750,6 +786,7 @@ cli_copy_db_getopts(int argc, char **argv)
 	/* install default values */
 	options.tableJobs = DEFAULT_TABLE_JOBS;
 	options.indexJobs = DEFAULT_INDEX_JOBS;
+	options.restoreOptions.jobs = DEFAULT_RESTORE_JOBS;
 	options.lObjectJobs = DEFAULT_LARGE_OBJECTS_JOBS;
 	options.splitTablesLargerThan.bytes = DEFAULT_SPLIT_TABLES_LARGER_THAN;
 
@@ -761,7 +798,7 @@ cli_copy_db_getopts(int argc, char **argv)
 	}
 
 	while ((c = getopt_long(argc, argv,
-							"S:T:D:J:I:b:L:cOBemlirRCN:xXCtfo:p:s:E:F:Q:iVvdzqh",
+							"S:T:D:J:I:b:L:cAPOXj:xBeMlUF:F:Q:irRCN:fp:ws:o:tE:Vvdzqh",
 							long_options, &option_index)) != -1)
 	{
 		switch (c)
@@ -900,6 +937,19 @@ cli_copy_db_getopts(int argc, char **argv)
 				break;
 			}
 
+			case 'j':
+			{
+				if (!stringToInt(optarg, &options.restoreOptions.jobs) ||
+					options.restoreOptions.jobs < 1 ||
+					options.restoreOptions.jobs > 128)
+				{
+					log_fatal("Failed to parse --restore-jobs count: \"%s\"", optarg);
+					++errors;
+				}
+				log_trace("--restore-jobs %d", options.restoreOptions.jobs);
+				break;
+			}
+
 			case 'B':
 			{
 				options.skipLargeObjects = true;
@@ -1009,6 +1059,13 @@ cli_copy_db_getopts(int argc, char **argv)
 				break;
 			}
 
+			case 'w':
+			{
+				options.slot.wal2jsonNumericAsString = true;
+				log_trace("--wal2json-numeric-as-string");
+				break;
+			}
+
 			case 'o':
 			{
 				strlcpy(options.origin, optarg, NAMEDATALEN);
@@ -1125,6 +1182,7 @@ cli_copy_db_getopts(int argc, char **argv)
 			}
 
 			case '?':
+			default:
 			{
 				commandline_help(stderr);
 				exit(EXIT_CODE_BAD_ARGS);
@@ -1133,6 +1191,13 @@ cli_copy_db_getopts(int argc, char **argv)
 		}
 	}
 
+	/* if we haven't set restore-jobs, set it to index-jobs */
+	if (options.restoreOptions.jobs == DEFAULT_RESTORE_JOBS)
+	{
+		options.restoreOptions.jobs = options.indexJobs;
+		log_trace("--restore-jobs %d", options.indexJobs);
+	}
+
 	if (options.connStrings.source_pguri == NULL ||
 		options.connStrings.target_pguri == NULL)
 	{
@@ -1140,6 +1205,14 @@ cli_copy_db_getopts(int argc, char **argv)
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
+	if (options.slot.wal2jsonNumericAsString &&
+		options.slot.plugin != STREAM_PLUGIN_WAL2JSON)
+	{
+		log_fatal("Option --wal2json-numeric-as-string "
+				  "requires option --plugin=wal2json");
+		exit(EXIT_CODE_BAD_ARGS);
+	}
+
 	if (!cli_prepare_pguris(&(options.connStrings)))
 	{
 		/* errors have already been logged */
