@@ -39,6 +39,7 @@ static CommandLine copy_db_command =
 		"  --dir                 Work directory to use\n"
 		"  --table-jobs          Number of concurrent COPY jobs to run\n"
 		"  --index-jobs          Number of concurrent CREATE INDEX jobs to run\n"
+		"  --restore-jobs        Number of concurrent jobs for pg_restore\n"
 		"  --drop-if-exists      On the target database, clean-up from a previous run first\n"
 		"  --roles               Also copy roles found on source to target\n"
 		"  --no-owner            Do not set ownership of objects to match the original database\n"
@@ -108,6 +109,7 @@ static CommandLine copy_data_command =
 		"  --dir                 Work directory to use\n"
 		"  --table-jobs          Number of concurrent COPY jobs to run\n"
 		"  --index-jobs          Number of concurrent CREATE INDEX jobs to run\n"
+		"  --restore-jobs        Number of concurrent jobs for pg_restore\n"
 		"  --skip-large-objects  Skip copying large objects (blobs)\n"
 		"  --filters <filename>  Use the filters defined in <filename>\n"
 		"  --restart             Allow restarting when temp files exist already\n"
@@ -176,6 +178,7 @@ static CommandLine copy_indexes_command =
 		"  --target             Postgres URI to the target database\n"
 		"  --dir                Work directory to use\n"
 		"  --index-jobs         Number of concurrent CREATE INDEX jobs to run\n"
+		"  --restore-jobs       Number of concurrent jobs for pg_restore\n"
 		"  --filters <filename> Use the filters defined in <filename>\n"
 		"  --restart            Allow restarting when temp files exist already\n"
 		"  --resume             Allow resuming operations after a failure\n"
@@ -228,13 +231,6 @@ cli_copy_schema(int argc, char **argv)
 
 	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_SCHEMA);
 
-	Summary summary = { 0 };
-	TopLevelTimings *timings = &(summary.timings);
-
-	(void) summary_set_current_time(timings, TIMING_STEP_START);
-
-	(void) summary_set_current_time(timings, TIMING_STEP_BEFORE_SCHEMA_DUMP);
-
 	/*
 	 * First, we need to open a snapshot that we're going to re-use in all our
 	 * connections to the source database. When the --snapshot option has been
@@ -266,8 +262,6 @@ cli_copy_schema(int argc, char **argv)
 	/* now close the snapshot we kept for the whole operation */
 	(void) copydb_close_snapshot(&copySpecs);
 
-	(void) summary_set_current_time(timings, TIMING_STEP_BEFORE_PREPARE_SCHEMA);
-
 	if (!copydb_target_prepare_schema(&copySpecs))
 	{
 		/* errors have already been logged */
@@ -275,20 +269,11 @@ cli_copy_schema(int argc, char **argv)
 		exit(EXIT_CODE_TARGET);
 	}
 
-	(void) summary_set_current_time(timings, TIMING_STEP_AFTER_PREPARE_SCHEMA);
-
-	(void) summary_set_current_time(timings, TIMING_STEP_BEFORE_FINALIZE_SCHEMA);
-
 	if (!copydb_target_finalize_schema(&copySpecs))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_TARGET);
 	}
-
-	(void) summary_set_current_time(timings, TIMING_STEP_AFTER_FINALIZE_SCHEMA);
-	(void) summary_set_current_time(timings, TIMING_STEP_END);
-
-	(void) print_summary(&summary, &copySpecs);
 }
 
 
@@ -307,11 +292,6 @@ cli_copy_data(int argc, char **argv)
 
 	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_ALL);
 
-	Summary summary = { 0 };
-	TopLevelTimings *timings = &(summary.timings);
-
-	(void) summary_set_current_time(timings, TIMING_STEP_START);
-
 	/*
 	 * First, we need to open a snapshot that we're going to re-use in all our
 	 * connections to the source database. When the --snapshot option has been
@@ -348,9 +328,6 @@ cli_copy_data(int argc, char **argv)
 				  copySpecs.sourceSnapshot.pguri);
 		exit(EXIT_CODE_SOURCE);
 	}
-
-	(void) summary_set_current_time(timings, TIMING_STEP_END);
-	(void) print_summary(&summary, &copySpecs);
 }
 
 
@@ -366,11 +343,6 @@ cli_copy_table_data(int argc, char **argv)
 
 	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_TABLE_DATA);
 
-	Summary summary = { 0 };
-	TopLevelTimings *timings = &(summary.timings);
-
-	(void) summary_set_current_time(timings, TIMING_STEP_START);
-
 	/*
 	 * First, we need to open a snapshot that we're going to re-use in all our
 	 * connections to the source database. When the --snapshot option has been
@@ -392,7 +364,7 @@ cli_copy_table_data(int argc, char **argv)
 		exit(EXIT_CODE_TARGET);
 	}
 
-	if (!copydb_copy_all_table_data(&copySpecs))
+	if (!copydb_copy_supervisor(&copySpecs))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
@@ -405,9 +377,6 @@ cli_copy_table_data(int argc, char **argv)
 				  copySpecs.sourceSnapshot.pguri);
 		exit(EXIT_CODE_SOURCE);
 	}
-
-	(void) summary_set_current_time(timings, TIMING_STEP_END);
-	(void) print_summary(&summary, &copySpecs);
 }
 
 
@@ -425,11 +394,6 @@ cli_copy_sequences(int argc, char **argv)
 
 	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_SET_SEQUENCES);
 
-	Summary summary = { 0 };
-	TopLevelTimings *timings = &(summary.timings);
-
-	(void) summary_set_current_time(timings, TIMING_STEP_START);
-
 	/*
 	 * First, we need to open a snapshot that we're going to re-use in all our
 	 * connections to the source database. When the --snapshot option has been
@@ -462,9 +426,6 @@ cli_copy_sequences(int argc, char **argv)
 				  copySpecs.sourceSnapshot.pguri);
 		exit(EXIT_CODE_SOURCE);
 	}
-
-	(void) summary_set_current_time(timings, TIMING_STEP_END);
-	(void) print_summary(&summary, &copySpecs);
 }
 
 
@@ -479,11 +440,6 @@ cli_copy_indexes(int argc, char **argv)
 
 	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_INDEXES);
 
-	Summary summary = { 0 };
-	TopLevelTimings *timings = &(summary.timings);
-
-	(void) summary_set_current_time(timings, TIMING_STEP_START);
-
 	/*
 	 * First, we need to open a snapshot that we're going to re-use in all our
 	 * connections to the source database. When the --snapshot option has been
@@ -516,9 +472,6 @@ cli_copy_indexes(int argc, char **argv)
 				  copySpecs.sourceSnapshot.pguri);
 		exit(EXIT_CODE_SOURCE);
 	}
-
-	(void) summary_set_current_time(timings, TIMING_STEP_END);
-	(void) print_summary(&summary, &copySpecs);
 }
 
 
@@ -534,11 +487,6 @@ cli_copy_constraints(int argc, char **argv)
 
 	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_CONSTRAINTS);
 
-	Summary summary = { 0 };
-	TopLevelTimings *timings = &(summary.timings);
-
-	(void) summary_set_current_time(timings, TIMING_STEP_START);
-
 	log_info("Create constraints");
 
 	/*
@@ -573,9 +521,6 @@ cli_copy_constraints(int argc, char **argv)
 				  copySpecs.sourceSnapshot.pguri);
 		exit(EXIT_CODE_SOURCE);
 	}
-
-	(void) summary_set_current_time(timings, TIMING_STEP_END);
-	(void) print_summary(&summary, &copySpecs);
 }
 
 
@@ -593,11 +538,6 @@ cli_copy_blobs(int argc, char **argv)
 	/* ensure defaults */
 	copySpecs.skipLargeObjects = false;
 
-	Summary summary = { 0 };
-	TopLevelTimings *timings = &(summary.timings);
-
-	(void) summary_set_current_time(timings, TIMING_STEP_START);
-
 	log_info("Copy large objects");
 
 	/*
@@ -630,9 +570,6 @@ cli_copy_blobs(int argc, char **argv)
 				  copySpecs.sourceSnapshot.pguri);
 		exit(EXIT_CODE_SOURCE);
 	}
-
-	(void) summary_set_current_time(timings, TIMING_STEP_END);
-	(void) print_summary(&summary, &copySpecs);
 }
 
 
@@ -680,7 +617,7 @@ cli_copy_extensions(int argc, char **argv)
 {
 	CopyDataSpec copySpecs = { 0 };
 
-	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_EXTENSION);
+	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_EXTENSIONS);
 
 	if (!copydb_prepare_snapshot(&copySpecs))
 	{
