@@ -152,6 +152,12 @@ cli_ping_getopts(int argc, char **argv)
 				exit(EXIT_CODE_QUIT);
 				break;
 			}
+
+			case '?':
+			default:
+			{
+				++errors;
+			}
 		}
 	}
 
@@ -162,6 +168,12 @@ cli_ping_getopts(int argc, char **argv)
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
+	if (errors > 0)
+	{
+		commandline_help(stderr);
+		exit(EXIT_CODE_BAD_ARGS);
+	}
+
 	/* prepare safe versions of the connection strings (without password) */
 	if (!cli_prepare_pguris(&(options.connStrings)))
 	{
@@ -224,7 +236,7 @@ cli_ping(int argc, char **argv)
 
 			if (!pgsql_set_gucs(&src, settings))
 			{
-				log_fatal("Failed to set our GUC settings on the target connection, "
+				log_fatal("Failed to set our GUC settings on the source connection, "
 						  "see above for details");
 				pgsql_finish(&src);
 				exit(EXIT_CODE_TARGET);
