@@ -8,6 +8,7 @@
 #include <getopt.h>
 #include <inttypes.h>
 
+#include "catalog.h"
 #include "cli_common.h"
 #include "cli_root.h"
 #include "copydb.h"
@@ -38,6 +39,7 @@ static CommandLine restore_schema_command =
 		"  --source             Postgres URI to the source database\n"
 		"  --target             Postgres URI to the target database\n"
 		"  --dir                Work directory to use\n"
+		"  --restore-jobs       Number of concurrent jobs for pg_restore\n"
 		"  --drop-if-exists     On the target database, clean-up from a previous run first\n"
 		"  --no-owner           Do not set ownership of objects to match the original database\n"
 		"  --no-acl             Prevent restoration of access privileges (grant/revoke commands).\n"
@@ -57,10 +59,13 @@ static CommandLine restore_schema_pre_data_command =
 		"  --source             Postgres URI to the source database\n"
 		"  --target             Postgres URI to the target database\n"
 		"  --dir                Work directory to use\n"
+		"  --restore-jobs       Number of concurrent jobs for pg_restore\n"
 		"  --drop-if-exists     On the target database, clean-up from a previous run first\n"
 		"  --no-owner           Do not set ownership of objects to match the original database\n"
 		"  --no-acl             Prevent restoration of access privileges (grant/revoke commands).\n"
 		"  --no-comments        Do not output commands to restore comments\n"
+		"  --skip-extensions    Skip restoring extensions\n" \
+		"  --skip-ext-comments  Skip restoring COMMENT ON EXTENSION\n" \
 		"  --filters <filename> Use the filters defined in <filename>\n"
 		"  --restart            Allow restarting when temp files exist already\n"
 		"  --resume             Allow resuming operations after a failure\n"
@@ -76,9 +81,12 @@ static CommandLine restore_schema_post_data_command =
 		"  --source             Postgres URI to the source database\n"
 		"  --target             Postgres URI to the target database\n"
 		"  --dir                Work directory to use\n"
+		"  --restore-jobs       Number of concurrent jobs for pg_restore\n"
 		"  --no-owner           Do not set ownership of objects to match the original database\n"
 		"  --no-acl             Prevent restoration of access privileges (grant/revoke commands).\n"
 		"  --no-comments        Do not output commands to restore comments\n"
+		"  --skip-extensions    Skip restoring extensions\n" \
+		"  --skip-ext-comments  Skip restoring COMMENT ON EXTENSION\n" \
 		"  --filters <filename> Use the filters defined in <filename>\n"
 		"  --restart            Allow restarting when temp files exist already\n"
 		"  --resume             Allow resuming operations after a failure\n"
@@ -93,7 +101,8 @@ static CommandLine restore_roles_command =
 		" --dir <dir> [ --source <URI> ] --target <URI> ",
 		"  --source             Postgres URI to the source database\n"
 		"  --target             Postgres URI to the target database\n"
-		"  --dir                Work directory to use\n",
+		"  --dir                Work directory to use\n"
+		"  --restore-jobs       Number of concurrent jobs for pg_restore\n",
 		cli_restore_schema_getopts,
 		cli_restore_roles);
 
@@ -144,10 +153,10 @@ cli_restore_schema_getopts(int argc, char **argv)
 		{ "source", required_argument, NULL, 'S' },
 		{ "target", required_argument, NULL, 'T' },
 		{ "dir", required_argument, NULL, 'D' },
-		{ "schema", required_argument, NULL, 's' },
 		{ "drop-if-exists", no_argument, NULL, 'c' }, /* pg_restore -c */
 		{ "no-owner", no_argument, NULL, 'O' },       /* pg_restore -O */
 		{ "no-comments", no_argument, NULL, 'X' },
+		{ "restore-jobs", required_argument, NULL, 'j' },      /* pg_restore --jobs */
 		{ "no-acl", no_argument, NULL, 'x' }, /* pg_restore -x */
 		{ "filter", required_argument, NULL, 'F' },
 		{ "filters", required_argument, NULL, 'F' },
@@ -170,6 +179,10 @@ cli_restore_schema_getopts(int argc, char **argv)
 
 	optind = 0;
 
+	/* install default values */
+	options.indexJobs = DEFAULT_INDEX_JOBS;
+	options.restoreOptions.jobs = DEFAULT_RESTORE_JOBS;
+
 	/* read values from the environment */
 	if (!cli_copydb_getenv(&options))
 	{
@@ -177,7 +190,7 @@ cli_restore_schema_getopts(int argc, char **argv)
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
-	while ((c = getopt_long(argc, argv, "S:T:D:s:cOxXFeErRCNVvdzqh",
+	while ((c = getopt_long(argc, argv, "S:T:D:cOXj:xF:eErRCN:Vvdzqh",
 							long_options, &option_index)) != -1)
 	{
 		switch (c)
@@ -229,6 +242,19 @@ cli_restore_schema_getopts(int argc, char **argv)
 				break;
 			}
 
+			case 'j':
+			{
+				if (!stringToInt(optarg, &options.restoreOptions.jobs) ||
+					options.restoreOptions.jobs < 1 ||
+					options.restoreOptions.jobs > 128)
+				{
+					log_fatal("Failed to parse --restore-jobs count: \"%s\"", optarg);
+					++errors;
+				}
+				log_trace("--restore-jobs %d", options.restoreOptions.jobs);
+				break;
+			}
+
 			case 'x':
 			{
 				options.restoreOptions.noACL = true;
@@ -364,6 +390,12 @@ cli_restore_schema_getopts(int argc, char **argv)
 				exit(EXIT_CODE_QUIT);
 				break;
 			}
+
+			case '?':
+			default:
+			{
+				++errors;
+			}
 		}
 	}
 
@@ -381,9 +413,17 @@ cli_restore_schema_getopts(int argc, char **argv)
 
 	if (errors > 0)
 	{
+		commandline_help(stderr);
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
+	/* if we haven't set restore-jobs, set it to index-jobs */
+	if (options.restoreOptions.jobs == DEFAULT_RESTORE_JOBS)
+	{
+		options.restoreOptions.jobs = options.indexJobs;
+		log_trace("--restore-jobs %d", options.indexJobs);
+	}
+
 	/* publish our option parsing in the global variable */
 	restoreDBoptions = options;
 
@@ -401,6 +441,13 @@ cli_restore_schema(int argc, char **argv)
 
 	(void) cli_restore_prepare_specs(&copySpecs);
 
+	/* we need access to the catalogs to filter the pg_restore --list */
+	if (!catalog_init_from_specs(&copySpecs))
+	{
+		log_error("Failed to initialize pgcopydb internal catalogs");
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
 	if (!copydb_target_prepare_schema(&copySpecs))
 	{
 		/* errors have already been logged */
@@ -412,6 +459,12 @@ cli_restore_schema(int argc, char **argv)
 		/* errors have already been logged */
 		exit(EXIT_CODE_TARGET);
 	}
+
+	if (!catalog_close_from_specs(&copySpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
 }
 
 
@@ -425,11 +478,24 @@ cli_restore_schema_pre_data(int argc, char **argv)
 
 	(void) cli_restore_prepare_specs(&copySpecs);
 
+	/* we need access to the catalogs to filter the pg_restore --list */
+	if (!catalog_init_from_specs(&copySpecs))
+	{
+		log_error("Failed to initialize pgcopydb internal catalogs");
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
 	if (!copydb_target_prepare_schema(&copySpecs))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_TARGET);
 	}
+
+	if (!catalog_close_from_specs(&copySpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
 }
 
 
@@ -443,11 +509,24 @@ cli_restore_schema_post_data(int argc, char **argv)
 
 	(void) cli_restore_prepare_specs(&copySpecs);
 
+	/* we need access to the catalogs to filter the pg_restore --list */
+	if (!catalog_init_from_specs(&copySpecs))
+	{
+		log_error("Failed to initialize pgcopydb internal catalogs");
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
 	if (!copydb_target_finalize_schema(&copySpecs))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_TARGET);
 	}
+
+	if (!catalog_close_from_specs(&copySpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
 }
 
 
@@ -613,6 +692,20 @@ cli_restore_prepare_specs(CopyDataSpec *copySpecs)
 		}
 	}
 
+	/*
+	 * Prepare our internal catalogs for storing the source database catalog
+	 * query results.
+	 */
+	copySpecs->section = DATA_SECTION_ALL;
+
+	if (!copydb_fetch_schema_and_prepare_specs(copySpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_TARGET);
+	}
+
+	copySpecs->section = DATA_SECTION_NONE;
+
 	log_info("Using pg_restore for Postgres \"%s\" at \"%s\"",
 			 pgPaths->pg_version,
 			 pgPaths->pg_restore);
