@@ -26,9 +26,6 @@ extern CommandLine version;
 extern CommandLine root;
 extern CommandLine *root_subcommands[];
 
-extern CommandLine root_with_debug;
-extern CommandLine *root_subcommands_with_debug[];
-
 int root_options(int argc, char **argv);
 
 /* cli_clone_follow.c */
@@ -36,14 +33,13 @@ void cli_clone(int argc, char **argv);
 void cli_follow(int argc, char **argv);
 
 /* cli_copy.h */
-extern CommandLine copy__db_command;
 extern CommandLine clone_command;
 extern CommandLine fork_command;
 extern CommandLine follow_command;
 extern CommandLine copy_commands;
 
 /* cli_snapshot.c */
-extern CommandLine create_snapshot_command;
+extern CommandLine snapshot_command;
 
 /* cli_dump.c */
 extern CommandLine dump_commands;
