@@ -22,33 +22,24 @@
 CopyDBOptions sentinelDBoptions = { 0 };
 
 static int cli_sentinel_getopts(int argc, char **argv);
-static void cli_sentinel_create(int argc, char **argv);
-static void cli_sentinel_drop(int argc, char **argv);
+static void cli_sentinel_setup(int argc, char **argv);
 static void cli_sentinel_set_startpos(int argc, char **argv);
 static void cli_sentinel_set_endpos(int argc, char **argv);
 static void cli_sentinel_set_apply(int argc, char **argv);
 static void cli_sentinel_set_prefetch(int argc, char **argv);
 static void cli_sentinel_get(int argc, char **argv);
 
-CommandLine sentinel_create_command =
+static bool cli_sentinel_init_specs(CopyDataSpec *copySpecs);
+
+CommandLine sentinel_setup_command =
 	make_command(
-		"create",
-		"Create the sentinel table on the source database",
-		" --source ... ",
-		"  --source      Postgres URI to the source database\n"
+		"setup",
+		"Setup the sentinel table",
+		"",
 		"  --startpos    Start replaying changes when reaching this LSN\n"
 		"  --endpos      Stop replaying changes when reaching this LSN\n",
 		cli_sentinel_getopts,
-		cli_sentinel_create);
-
-CommandLine sentinel_drop_command =
-	make_command(
-		"drop",
-		"Drop the sentinel table on the source database",
-		" --source ... ",
-		"  --source      Postgres URI to the source database\n",
-		cli_sentinel_getopts,
-		cli_sentinel_drop);
+		cli_sentinel_setup);
 
 CommandLine sentinel_get_command =
 	make_command(
@@ -111,8 +102,7 @@ static CommandLine sentinel_set_commands =
 					 NULL, NULL, NULL, sentinel_set_subcommands);
 
 static CommandLine *sentinel_subcommands[] = {
-	&sentinel_create_command,
-	&sentinel_drop_command,
+	&sentinel_setup_command,
 	&sentinel_get_command,
 	&sentinel_set_commands,
 	NULL
@@ -136,6 +126,7 @@ cli_sentinel_getopts(int argc, char **argv)
 
 	static struct option long_options[] = {
 		{ "source", required_argument, NULL, 'S' },
+		{ "dir", required_argument, NULL, 'D' },
 		{ "endpos", required_argument, NULL, 'E' },
 		{ "startpos", required_argument, NULL, 's' },
 		{ "current", no_argument, NULL, 'C' },
@@ -159,7 +150,7 @@ cli_sentinel_getopts(int argc, char **argv)
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
-	while ((c = getopt_long(argc, argv, "S:s:E:CVvdzqh",
+	while ((c = getopt_long(argc, argv, "S:D:E:s:CJVvdzqh",
 							long_options, &option_index)) != -1)
 	{
 		switch (c)
@@ -177,6 +168,13 @@ cli_sentinel_getopts(int argc, char **argv)
 				break;
 			}
 
+			case 'D':
+			{
+				strlcpy(options.dir, optarg, MAXPGPATH);
+				log_trace("--dir %s", options.dir);
+				break;
+			}
+
 			case 'E':
 			{
 				if (!parseLSN(optarg, &(options.endpos)))
@@ -283,30 +281,40 @@ cli_sentinel_getopts(int argc, char **argv)
 				exit(EXIT_CODE_QUIT);
 				break;
 			}
+
+			case '?':
+			default:
+			{
+				++errors;
+			}
 		}
 	}
 
-	if (options.connStrings.source_pguri == NULL)
+	if (options.currentpos)
 	{
-		log_fatal("Options --source is mandatory");
-		++errors;
-	}
+		if (options.endpos != InvalidXLogRecPtr)
+		{
+			log_fatal("Please choose only one of --endpos and --current");
+			++errors;
+		}
 
-	/* prepare safe versions of the connection strings (without password) */
-	if (!cli_prepare_pguris(&(options.connStrings)))
-	{
-		/* errors have already been logged */
-		exit(EXIT_CODE_INTERNAL_ERROR);
-	}
+		if (options.connStrings.source_pguri == NULL)
+		{
+			log_fatal("Options --source is mandatory when using --current");
+			++errors;
+		}
 
-	if (options.currentpos && options.endpos != InvalidXLogRecPtr)
-	{
-		log_fatal("Please choose only one of --endpos and --current");
-		++errors;
+		/* prepare safe versions of the connection strings (without password) */
+		if (!cli_prepare_pguris(&(options.connStrings)))
+		{
+			/* errors have already been logged */
+			++errors;
+		}
 	}
 
 	if (errors > 0)
 	{
+		commandline_help(stderr);
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
@@ -318,63 +326,33 @@ cli_sentinel_getopts(int argc, char **argv)
 
 
 /*
- * cli_sentinel_create creates the pgcopydb.sentinel table on the source
- * database.
+ * cli_sentinel_setup sets-up the sentinel table in pgcopydb catalogs.
  */
 static void
-cli_sentinel_create(int argc, char **argv)
+cli_sentinel_setup(int argc, char **argv)
 {
-	CopyDataSpec copySpecs = { 0 };
-
 	if (argc > 0)
 	{
 		commandline_help(stderr);
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
-	(void) find_pg_commands(&(copySpecs.pgPaths));
+	CopyDataSpec copySpecs = { 0 };
 
-	if (!copydb_init_specs(&copySpecs, &sentinelDBoptions, DATA_SECTION_ALL))
+	if (!cli_sentinel_init_specs(&copySpecs))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	if (!stream_create_sentinel(&copySpecs,
-								sentinelDBoptions.startpos,
-								sentinelDBoptions.endpos))
+	DatabaseCatalog *sourceDB = &(copySpecs.catalogs.source);
+
+	if (!sentinel_setup(sourceDB,
+						sentinelDBoptions.startpos,
+						sentinelDBoptions.endpos))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
-	}
-}
-
-
-/*
- * cli_sentinel_drop drops the pgcopydb.sentinel table on the source database.
- */
-static void
-cli_sentinel_drop(int argc, char **argv)
-{
-	char *pguri = (char *) sentinelDBoptions.connStrings.source_pguri;
-	PGSQL pgsql = { 0 };
-
-	if (argc > 0)
-	{
-		commandline_help(stderr);
-		exit(EXIT_CODE_BAD_ARGS);
-	}
-
-	if (!pgsql_init(&pgsql, pguri, PGSQL_CONN_SOURCE))
-	{
-		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
-	}
-
-	if (!pgsql_execute(&pgsql, "drop schema pgcopydb cascade"))
-	{
-		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 }
 
@@ -422,19 +400,20 @@ cli_sentinel_set_startpos(int argc, char **argv)
 		}
 	}
 
-	char *pguri = (char *) sentinelDBoptions.connStrings.source_pguri;
-	PGSQL pgsql = { 0 };
+	CopyDataSpec copySpecs = { 0 };
 
-	if (!pgsql_init(&pgsql, pguri, PGSQL_CONN_SOURCE))
+	if (!cli_sentinel_init_specs(&copySpecs))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	if (!pgsql_update_sentinel_startpos(&pgsql, startpos))
+	DatabaseCatalog *sourceDB = &(copySpecs.catalogs.source);
+
+	if (!sentinel_update_startpos(sourceDB, startpos))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 }
 
@@ -491,54 +470,47 @@ cli_sentinel_set_endpos(int argc, char **argv)
 		}
 	}
 
-	char *pguri = (char *) sentinelDBoptions.connStrings.source_pguri;
-	PGSQL pgsql = { 0 };
+	CopyDataSpec copySpecs = { 0 };
 
-	if (!pgsql_init(&pgsql, pguri, PGSQL_CONN_SOURCE))
+	if (!cli_sentinel_init_specs(&copySpecs))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
-	}
-
-	if (!pgsql_begin(&pgsql))
-	{
-		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
-	}
-
-	if (!pgsql_server_version(&pgsql))
-	{
-		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
-	}
-
-	if (!pgsql_update_sentinel_endpos(&pgsql, useCurrentLSN, endpos))
-	{
-		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
-	}
-
-	if (!pgsql_commit(&pgsql))
-	{
-		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
 	if (useCurrentLSN)
 	{
-		CopyDBSentinel sentinel = { 0 };
+		char *pguri = (char *) sentinelDBoptions.connStrings.source_pguri;
 
-		if (!pgsql_get_sentinel(&pgsql, &sentinel))
+		if (!stream_fetch_current_lsn(&endpos, pguri, PGSQL_CONN_SOURCE))
 		{
-			/* errors have already been logged */
 			exit(EXIT_CODE_SOURCE);
 		}
 
-		log_info("pgcopydb sentinel endpos has been set to %X/%X",
-				 LSN_FORMAT_ARGS(sentinel.endpos));
-
-		fformat(stdout, "%X/%X\n", LSN_FORMAT_ARGS(sentinel.endpos));
+		log_info("Fetched endpos %X/%X from source database",
+				 LSN_FORMAT_ARGS(endpos));
 	}
+
+	DatabaseCatalog *sourceDB = &(copySpecs.catalogs.source);
+
+	if (!sentinel_update_endpos(sourceDB, endpos))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	CopyDBSentinel sentinel = { 0 };
+
+	if (!sentinel_get(sourceDB, &sentinel))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	log_info("pgcopydb sentinel endpos has been set to %X/%X",
+			 LSN_FORMAT_ARGS(sentinel.endpos));
+
+	fformat(stdout, "%X/%X\n", LSN_FORMAT_ARGS(sentinel.endpos));
 }
 
 
@@ -550,25 +522,26 @@ cli_sentinel_set_endpos(int argc, char **argv)
 static void
 cli_sentinel_set_apply(int argc, char **argv)
 {
-	char *pguri = (char *) sentinelDBoptions.connStrings.source_pguri;
-	PGSQL pgsql = { 0 };
-
 	if (argc > 0)
 	{
 		commandline_help(stderr);
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
-	if (!pgsql_init(&pgsql, pguri, PGSQL_CONN_SOURCE))
+	CopyDataSpec copySpecs = { 0 };
+
+	if (!cli_sentinel_init_specs(&copySpecs))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	if (!pgsql_update_sentinel_apply(&pgsql, true))
+	DatabaseCatalog *sourceDB = &(copySpecs.catalogs.source);
+
+	if (!sentinel_update_apply(sourceDB, true))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 }
 
@@ -581,25 +554,26 @@ cli_sentinel_set_apply(int argc, char **argv)
 static void
 cli_sentinel_set_prefetch(int argc, char **argv)
 {
-	char *pguri = (char *) sentinelDBoptions.connStrings.source_pguri;
-	PGSQL pgsql = { 0 };
-
 	if (argc > 0)
 	{
 		commandline_help(stderr);
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
-	if (!pgsql_init(&pgsql, pguri, PGSQL_CONN_SOURCE))
+	CopyDataSpec copySpecs = { 0 };
+
+	if (!cli_sentinel_init_specs(&copySpecs))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	if (!pgsql_update_sentinel_apply(&pgsql, false))
+	DatabaseCatalog *sourceDB = &(copySpecs.catalogs.source);
+
+	if (!sentinel_update_apply(sourceDB, false))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 }
 
@@ -610,24 +584,25 @@ cli_sentinel_set_prefetch(int argc, char **argv)
 static void
 cli_sentinel_get(int argc, char **argv)
 {
-	char *pguri = (char *) sentinelDBoptions.connStrings.source_pguri;
-	PGSQL pgsql = { 0 };
-
 	if (argc > 0)
 	{
 		commandline_help(stderr);
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
-	if (!pgsql_init(&pgsql, pguri, PGSQL_CONN_SOURCE))
+	CopyDataSpec copySpecs = { 0 };
+
+	if (!cli_sentinel_init_specs(&copySpecs))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
+	DatabaseCatalog *sourceDB = &(copySpecs.catalogs.source);
+
 	CopyDBSentinel sentinel = { 0 };
 
-	if (!pgsql_get_sentinel(&pgsql, &sentinel))
+	if (!sentinel_get(sourceDB, &sentinel))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_SOURCE);
@@ -685,3 +660,52 @@ cli_sentinel_get(int argc, char **argv)
 				LSN_FORMAT_ARGS(sentinel.replay_lsn));
 	}
 }
+
+
+/*
+ * cli_sentinel_init_specs initializes our CopyDataSpec from cli options.
+ */
+static bool
+cli_sentinel_init_specs(CopyDataSpec *copySpecs)
+{
+	char *dir =
+		IS_EMPTY_STRING_BUFFER(sentinelDBoptions.dir)
+		? NULL
+		: sentinelDBoptions.dir;
+
+	bool service = false;
+	char *serviceName = NULL;
+	bool createWorkDir = false;
+
+	/* pretend --resume, allowing to work on an existing directory */
+	bool restart = false;
+	bool resume = true;
+
+	sentinelDBoptions.notConsistent = true;
+
+	if (!copydb_init_workdir(copySpecs,
+							 dir,
+							 service,
+							 serviceName,
+							 restart,
+							 resume,
+							 createWorkDir))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	if (!copydb_init_specs(copySpecs, &sentinelDBoptions, DATA_SECTION_NONE))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	if (!catalog_init_from_specs(copySpecs))
+	{
+		log_error("Failed to initialize pgcopydb internal catalogs");
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	return true;
+}
