@@ -24,16 +24,17 @@
 static int cli_create_snapshot_getopts(int argc, char **argv);
 static void cli_create_snapshot(int argc, char **argv);
 
-CommandLine create_snapshot_command =
+CommandLine snapshot_command =
 	make_command(
 		"snapshot",
 		"Create and export a snapshot on the source database",
 		" --source ... ",
-		"  --source         Postgres URI to the source database\n"
-		"  --dir            Work directory to use\n"
-		"  --follow         Implement logical decoding to replay changes\n"
-		"  --plugin         Output plugin to use (test_decoding, wal2json)\n"
-		"  --slot-name      Use this Postgres replication slot name\n",
+		"  --source                      Postgres URI to the source database\n"
+		"  --dir                         Work directory to use\n"
+		"  --follow                      Implement logical decoding to replay changes\n"
+		"  --plugin                      Output plugin to use (test_decoding, wal2json)\n"
+		"  --wal2json-numeric-as-string  Print numeric data type as string when using wal2json output plugin\n"
+		"  --slot-name                   Use this Postgres replication slot name\n",
 		cli_create_snapshot_getopts,
 		cli_create_snapshot);
 
@@ -51,6 +52,7 @@ cli_create_snapshot_getopts(int argc, char **argv)
 		{ "dir", required_argument, NULL, 'D' },
 		{ "follow", no_argument, NULL, 'f' },
 		{ "plugin", required_argument, NULL, 'p' },
+		{ "wal2json-numeric-as-string", no_argument, NULL, 'w' },
 		{ "slot-name", required_argument, NULL, 's' },
 		{ "version", no_argument, NULL, 'V' },
 		{ "verbose", no_argument, NULL, 'v' },
@@ -71,7 +73,7 @@ cli_create_snapshot_getopts(int argc, char **argv)
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
-	while ((c = getopt_long(argc, argv, "S:T:D:Vvdzqh",
+	while ((c = getopt_long(argc, argv, "S:D:fp:ws:Vvdzqh",
 							long_options, &option_index)) != -1)
 	{
 		switch (c)
@@ -118,6 +120,13 @@ cli_create_snapshot_getopts(int argc, char **argv)
 				break;
 			}
 
+			case 'w':
+			{
+				options.slot.wal2jsonNumericAsString = true;
+				log_trace("--wal2json-numeric-as-string");
+				break;
+			}
+
 			case 'V':
 			{
 				/* keeper_cli_print_version prints version and exits. */
@@ -183,6 +192,13 @@ cli_create_snapshot_getopts(int argc, char **argv)
 				exit(EXIT_CODE_QUIT);
 				break;
 			}
+
+			case '?':
+			default:
+			{
+				++errors;
+				break;
+			}
 		}
 	}
 
@@ -192,6 +208,14 @@ cli_create_snapshot_getopts(int argc, char **argv)
 		++errors;
 	}
 
+	if (options.slot.wal2jsonNumericAsString &&
+		options.slot.plugin != STREAM_PLUGIN_WAL2JSON)
+	{
+		log_fatal("Option --wal2json-numeric-as-string "
+				  "requires option --plugin=wal2json");
+		exit(EXIT_CODE_BAD_ARGS);
+	}
+
 	/* prepare safe versions of the connection strings (without password) */
 	if (!cli_prepare_pguris(&(options.connStrings)))
 	{
@@ -218,6 +242,7 @@ cli_create_snapshot_getopts(int argc, char **argv)
 
 	if (errors > 0)
 	{
+		commandline_help(stderr);
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
@@ -297,7 +322,7 @@ cli_create_snapshot(int argc, char **argv)
 							   createSNoptions.origin,
 							   createSNoptions.endpos,
 							   STREAM_MODE_CATCHUP,
-							   &(copySpecs.catalog),
+							   &(copySpecs.catalogs.source),
 							   createSNoptions.stdIn,
 							   createSNoptions.stdOut,
 							   logSQL))
@@ -306,11 +331,17 @@ cli_create_snapshot(int argc, char **argv)
 			exit(EXIT_CODE_INTERNAL_ERROR);
 		}
 
-		char *logrep_pguri = streamSpecs.connStrings->logrep_pguri;
+		/*
+		 * Make sure to register our setup here, as usually the command
+		 * `pgcopydb snapshot` is used first.
+		 */
+		if (!catalog_register_setup_from_specs(&copySpecs))
+		{
+			/* errors have already been logged */
+			exit(EXIT_CODE_INTERNAL_ERROR);
+		}
 
-		if (!copydb_create_logical_replication_slot(&copySpecs,
-													logrep_pguri,
-													&(streamSpecs.slot)))
+		if (!follow_export_snapshot(&copySpecs, &streamSpecs))
 		{
 			/* errors have already been logged */
 			exit(EXIT_CODE_INTERNAL_ERROR);
