@@ -13,6 +13,7 @@
 #include "pqexpbuffer.h"
 #include "dumputils.h"
 
+#include "catalog.h"
 #include "copydb.h"
 #include "env_utils.h"
 #include "lock_utils.h"
@@ -24,6 +25,11 @@
 #include "summary.h"
 
 
+static bool copydb_append_table_hook(void *context, SourceTable *table);
+static bool copydb_copy_database_properties_hook(void *ctx,
+												 SourceProperty *property);
+
+
 /*
  * copydb_objectid_has_been_processed_already returns true when a doneFile
  * could be found on-disk for the given target object OID.
@@ -31,14 +37,17 @@
 bool
 copydb_objectid_has_been_processed_already(CopyDataSpec *specs, uint32_t oid)
 {
-	char doneFile[MAXPGPATH] = { 0 };
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
 
-	/* build the doneFile for the target index or constraint */
-	sformat(doneFile, sizeof(doneFile), "%s/%u.done",
-			specs->cfPaths.idxdir,
-			oid);
+	bool done = false;
 
-	return file_exists(doneFile);
+	if (!summary_lookup_oid(sourceDB, oid, &done))
+	{
+		/* errors have aleady been logged */
+		return false;
+	}
+
+	return done;
 }
 
 
@@ -51,73 +60,64 @@ copydb_dump_source_schema(CopyDataSpec *specs,
 						  const char *snapshot,
 						  PostgresDumpSection section)
 {
-	SourceExtensionArray *extensionArray = NULL;
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
 
-	if (specs->skipExtensions)
+	if (!summary_start_timing(sourceDB, TIMING_SECTION_DUMP_SCHEMA))
 	{
-		extensionArray = &(specs->catalog.extensionArray);
+		/* errors have already been logged */
+		return false;
 	}
 
 	if (section == PG_DUMP_SECTION_SCHEMA ||
 		section == PG_DUMP_SECTION_PRE_DATA ||
 		section == PG_DUMP_SECTION_ALL)
 	{
-		if (file_exists(specs->cfPaths.done.preDataDump))
+		if (file_exists(specs->dumpPaths.preFilename))
 		{
 			log_info("Skipping pg_dump --section=pre-data, "
 					 "as \"%s\" already exists",
-					 specs->cfPaths.done.preDataDump);
+					 specs->dumpPaths.preFilename);
 		}
 		else if (!pg_dump_db(&(specs->pgPaths),
 							 &(specs->connStrings),
 							 snapshot,
 							 "pre-data",
 							 &(specs->filters),
-							 extensionArray,
+							 &(specs->catalogs.filter),
 							 specs->dumpPaths.preFilename))
 		{
 			/* errors have already been logged */
 			return false;
 		}
-
-		/* now write the doneFile to keep track */
-		if (!write_file("", 0, specs->cfPaths.done.preDataDump))
-		{
-			log_error("Failed to write the tracking file \"%s\"",
-					  specs->cfPaths.done.preDataDump);
-			return false;
-		}
 	}
 
 	if (section == PG_DUMP_SECTION_SCHEMA ||
 		section == PG_DUMP_SECTION_POST_DATA ||
 		section == PG_DUMP_SECTION_ALL)
 	{
-		if (file_exists(specs->cfPaths.done.postDataDump))
+		if (file_exists(specs->dumpPaths.postFilename))
 		{
 			log_info("Skipping pg_dump --section=post-data, "
 					 "as \"%s\" already exists",
-					 specs->cfPaths.done.postDataDump);
+					 specs->dumpPaths.postFilename);
 		}
 		else if (!pg_dump_db(&(specs->pgPaths),
 							 &(specs->connStrings),
 							 snapshot,
 							 "post-data",
 							 &(specs->filters),
-							 extensionArray,
+							 &(specs->catalogs.filter),
 							 specs->dumpPaths.postFilename))
 		{
 			/* errors have already been logged */
 			return false;
 		}
+	}
 
-		/* now write the doneFile to keep track */
-		if (!write_file("", 0, specs->cfPaths.done.postDataDump))
-		{
-			log_error("Failed to write the tracking file \"%s\"",
-					  specs->cfPaths.done.postDataDump);
-			return false;
-		}
+	if (!summary_stop_timing(sourceDB, TIMING_SECTION_DUMP_SCHEMA))
+	{
+		/* errors have already been logged */
+		return false;
 	}
 
 	return true;
@@ -131,15 +131,23 @@ copydb_dump_source_schema(CopyDataSpec *specs,
 bool
 copydb_target_prepare_schema(CopyDataSpec *specs)
 {
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
+
+	if (!summary_start_timing(sourceDB, TIMING_SECTION_PREPARE_SCHEMA))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	if (!file_exists(specs->dumpPaths.preFilename))
 	{
 		log_fatal("File \"%s\" does not exists", specs->dumpPaths.preFilename);
 		return false;
 	}
 
-	if (file_exists(specs->cfPaths.done.preDataRestore))
+	if (specs->runState.schemaPreDataHasBeenRestored)
 	{
-		log_info("Skipping pg_restore of pre-data section, "
+		log_info("Skipping pg_restore for pre-data section, "
 				 "done on a previous run");
 		return true;
 	}
@@ -161,7 +169,7 @@ copydb_target_prepare_schema(CopyDataSpec *specs)
 	{
 		log_error("Failed to prepare the pg_restore --use-list catalogs, "
 				  "see above for details");
-		return true;
+		return false;
 	}
 
 	/*
@@ -192,11 +200,9 @@ copydb_target_prepare_schema(CopyDataSpec *specs)
 		return false;
 	}
 
-	/* now write the doneFile to keep track */
-	if (!write_file("", 0, specs->cfPaths.done.preDataRestore))
+	if (!summary_stop_timing(sourceDB, TIMING_SECTION_PREPARE_SCHEMA))
 	{
-		log_error("Failed to write the tracking file \"%s\"",
-				  specs->cfPaths.done.preDataRestore);
+		/* errors have already been logged */
 		return false;
 	}
 
@@ -204,6 +210,13 @@ copydb_target_prepare_schema(CopyDataSpec *specs)
 }
 
 
+typedef struct CopyPropertiesContext
+{
+	CopyDataSpec *specs;
+	PGSQL *dst;
+} CopyPropertiesContext;
+
+
 /*
  * copydb_copy_database_properties uses ALTER DATABASE SET commands to set the
  * properties on the target database to look the same way as on the source
@@ -212,6 +225,8 @@ copydb_target_prepare_schema(CopyDataSpec *specs)
 bool
 copydb_copy_database_properties(CopyDataSpec *specs)
 {
+	const char *s_dbname = specs->connStrings.safeSourcePGURI.uriParams.dbname;
+
 	PGSQL dst = { 0 };
 
 	if (!pgsql_init(&dst, specs->connStrings.target_pguri, PGSQL_CONN_TARGET))
@@ -226,87 +241,21 @@ copydb_copy_database_properties(CopyDataSpec *specs)
 		return false;
 	}
 
-	PGconn *conn = dst.connection;
-	SourceRole *rolesHashByName = specs->targetCatalog.rolesHashByName;
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
 
-	if (specs->catalog.gucsArray.count > 0)
+	CopyPropertiesContext context = {
+		.specs = specs,
+		.dst = &dst
+	};
+
+	if (!catalog_iter_s_database_guc(sourceDB,
+									 s_dbname,
+									 &context,
+									 &copydb_copy_database_properties_hook))
 	{
-		for (int i = 0; i < specs->catalog.gucsArray.count; i++)
-		{
-			SourceProperty *property = &(specs->catalog.gucsArray.array[i]);
-
-			/*
-			 * ALTER ROLE rolname IN DATABASE datname SET ...
-			 */
-			if (property->roleInDatabase)
-			{
-				char *rolname = property->rolname;
-				int len = strlen(rolname);
-
-				SourceRole *role = NULL;
-
-				HASH_FIND(hh, rolesHashByName, rolname, len, role);
-
-				if (role != NULL)
-				{
-					PQExpBuffer command = createPQExpBuffer();
-
-					makeAlterConfigCommand(conn, property->setconfig,
-										   "ROLE", property->rolname,
-										   "DATABASE", specs->connStrings.safeTargetPGURI.uriParams.dbname,
-										   command);
-
-					/* chomp the \n */
-					if (command->data[command->len - 1] == '\n')
-					{
-						command->data[command->len - 1] = '\0';
-					}
-
-					log_info("%s", command->data);
-
-					if (!pgsql_execute(&dst, command->data))
-					{
-						/* errors have already been logged */
-						return false;
-					}
-
-					destroyPQExpBuffer(command);
-				}
-				else
-				{
-					log_warn("Skipping database properties for role %s which "
-							 "does not exists on the target database",
-							 rolname);
-				}
-			}
-
-			/*
-			 * ALTER DATABASE datname SET ...
-			 */
-			else
-			{
-				PQExpBuffer command = createPQExpBuffer();
-
-				makeAlterConfigCommand(conn, property->setconfig,
-									   "DATABASE", specs->connStrings.safeTargetPGURI.uriParams.dbname, NULL, NULL,
-									   command);
-
-				if (command->data[command->len - 1] == '\n')
-				{
-					command->data[command->len - 1] = '\0';
-				}
-
-				log_info("%s", command->data);
-
-				if (!pgsql_execute(&dst, command->data))
-				{
-					/* errors have already been logged */
-					return false;
-				}
-
-				destroyPQExpBuffer(command);
-			}
-		}
+		/* errors have already been logged */
+		(void) pgsql_rollback(&dst);
+		return false;
 	}
 
 	if (!pgsql_commit(&dst))
@@ -319,6 +268,118 @@ copydb_copy_database_properties(CopyDataSpec *specs)
 }
 
 
+/*
+ * copydb_copy_database_properties_hook is an iterator callback function.
+ */
+static bool
+copydb_copy_database_properties_hook(void *ctx, SourceProperty *property)
+{
+	CopyPropertiesContext *context = (CopyPropertiesContext *) ctx;
+
+	CopyDataSpec *specs = context->specs;
+
+	PGSQL *dst = context->dst;
+	PGconn *conn = dst->connection;
+
+	const char *t_dbname = specs->connStrings.safeTargetPGURI.uriParams.dbname;
+
+	/*
+	 * ALTER ROLE rolname IN DATABASE datname SET ...
+	 */
+	if (property->roleInDatabase)
+	{
+		DatabaseCatalog *targetDB = &(specs->catalogs.target);
+
+		SourceRole *role = (SourceRole *) calloc(1, sizeof(SourceRole));
+
+		if (role == NULL)
+		{
+			log_error(ALLOCATION_FAILED_ERROR);
+			return false;
+		}
+
+		if (!catalog_lookup_s_role_by_name(targetDB, property->rolname, role))
+		{
+			/* errors have already been logged */
+			free(role);
+			return false;
+		}
+
+		if (role->oid > 0)
+		{
+			PQExpBuffer command = createPQExpBuffer();
+
+			makeAlterConfigCommand(conn, property->setconfig,
+								   "ROLE", property->rolname,
+								   "DATABASE",
+								   t_dbname,
+								   command);
+
+			/* chomp the \n */
+			if (command->data[command->len - 1] == '\n')
+			{
+				command->data[command->len - 1] = '\0';
+			}
+
+			log_info("%s", command->data);
+
+			if (!pgsql_execute(dst, command->data))
+			{
+				/* errors have already been logged */
+				free(role);
+				return false;
+			}
+
+			destroyPQExpBuffer(command);
+		}
+		else
+		{
+			log_warn("Skipping database properties for role %s which "
+					 "does not exists on the target database",
+					 property->rolname);
+		}
+
+		free(role);
+	}
+
+	/*
+	 * ALTER DATABASE datname SET ...
+	 */
+	else
+	{
+		PQExpBuffer command = createPQExpBuffer();
+
+		makeAlterConfigCommand(conn, property->setconfig,
+							   "DATABASE", t_dbname,
+							   NULL, NULL,
+							   command);
+
+		if (command->data[command->len - 1] == '\n')
+		{
+			command->data[command->len - 1] = '\0';
+		}
+
+		log_info("%s", command->data);
+
+		if (!pgsql_execute(dst, command->data))
+		{
+			/* errors have already been logged */
+			return false;
+		}
+
+		destroyPQExpBuffer(command);
+	}
+
+	return true;
+}
+
+
+typedef struct DropTableContext
+{
+	PQExpBuffer query;
+	uint64_t tableIndex;
+} DropTableContext;
+
 /*
  * copydb_target_drop_tables prepares and executes a SQL query that prepares
  * our target database by means of a DROP IF EXISTS ... CASCADE statement that
@@ -327,28 +388,36 @@ copydb_copy_database_properties(CopyDataSpec *specs)
 bool
 copydb_target_drop_tables(CopyDataSpec *specs)
 {
-	log_info("Drop tables on the target database, per --drop-if-exists");
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
+	CatalogStats stats = { 0 };
 
-	SourceTableArray *tableArray = &(specs->catalog.sourceTableArray);
-
-	if (tableArray->count == 0)
+	if (!catalog_stats(sourceDB, &stats))
 	{
-		log_info("No tables to migrate, skipping drop tables on the target database");
+		/* errors have already been logged */
+		return false;
+	}
+
+	if (stats.count.tables == 0)
+	{
+		log_info("No tables to migrate, skipping drop tables "
+				 "on the target database");
 		return true;
 	}
 
+	log_info("Drop tables on the target database, per --drop-if-exists");
+
 	PQExpBuffer query = createPQExpBuffer();
 
+	DropTableContext context = { .query = query, .tableIndex = 0 };
+
 	appendPQExpBufferStr(query, "DROP TABLE IF EXISTS");
 
-	for (int tableIndex = 0; tableIndex < tableArray->count; tableIndex++)
+	if (!catalog_iter_s_table(sourceDB, &context, &copydb_append_table_hook))
 	{
-		SourceTable *source = &(tableArray->array[tableIndex]);
-
-		appendPQExpBuffer(query, "%s %s.%s",
-						  tableIndex == 0 ? " " : ",",
-						  source->nspname,
-						  source->relname);
+		log_error("Failed to create DROP IF EXISTS query: "
+				  "see above for details");
+		destroyPQExpBuffer(query);
+		return false;
 	}
 
 	appendPQExpBufferStr(query, " CASCADE");
@@ -383,6 +452,29 @@ copydb_target_drop_tables(CopyDataSpec *specs)
 }
 
 
+/*
+ * copydb_append_table_hook is an iterator callback function.
+ */
+static bool
+copydb_append_table_hook(void *ctx, SourceTable *table)
+{
+	if (table == NULL)
+	{
+		log_error("BUG: copydb_append_table_hook called with a NULL table");
+		return false;
+	}
+
+	DropTableContext *context = (DropTableContext *) ctx;
+
+	appendPQExpBuffer(context->query, "%s %s.%s",
+					  context->tableIndex++ == 0 ? " " : ",",
+					  table->nspname,
+					  table->relname);
+
+	return true;
+}
+
+
 /*
  * copydb_target_finalize_schema finalizes the schema after all the data has
  * been copied over, and after indexes and their constraints have been created
@@ -391,15 +483,23 @@ copydb_target_drop_tables(CopyDataSpec *specs)
 bool
 copydb_target_finalize_schema(CopyDataSpec *specs)
 {
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
+
+	if (!summary_start_timing(sourceDB, TIMING_SECTION_FINALIZE_SCHEMA))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	if (!file_exists(specs->dumpPaths.postFilename))
 	{
 		log_fatal("File \"%s\" does not exists", specs->dumpPaths.postFilename);
 		return false;
 	}
 
-	if (file_exists(specs->cfPaths.done.postDataRestore))
+	if (specs->runState.schemaPostDataHasBeenRestored)
 	{
-		log_info("Skipping pg_restore --section=pre-data, "
+		log_info("Skipping pg_restore for --section=post-data, "
 				 "done on a previous run");
 		return true;
 	}
@@ -422,11 +522,9 @@ copydb_target_finalize_schema(CopyDataSpec *specs)
 		return false;
 	}
 
-	/* now write the doneFile to keep track */
-	if (!write_file("", 0, specs->cfPaths.done.postDataRestore))
+	if (!summary_stop_timing(sourceDB, TIMING_SECTION_FINALIZE_SCHEMA))
 	{
-		log_error("Failed to write the tracking file \"%s\"",
-				  specs->cfPaths.done.postDataRestore);
+		/* errors have already been logged */
 		return false;
 	}
 
@@ -485,8 +583,8 @@ copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section)
 	 *
 	 * Here's how to filter out some objects with pg_restore:
 	 *
-	 *   1. pg_restore -f out.list --list post.dump
-	 *   2. edit post.list to comment out lines
+	 *   1. pg_restore -f post.list --list post.dump
+	 *   2. edit post.list to comment out lines and save as filtered.list
 	 *   3. pg_restore --use-list filtered.list post.dump
 	 */
 	ArchiveContentArray contents = { 0 };
