@@ -40,6 +40,16 @@ follow_export_snapshot(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs)
 		return false;
 	}
 
+	if (!catalog_setup_replication(streamSpecs->sourceDB,
+								   streamSpecs->slot.snapshot,
+								   OutputPluginToString(streamSpecs->slot.plugin),
+								   streamSpecs->slot.slotName))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+
 	return true;
 }
 
@@ -143,15 +153,9 @@ follow_reset_sequences(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs)
 bool
 follow_init_sentinel(StreamSpecs *specs, CopyDBSentinel *sentinel)
 {
-	PGSQL pgsql = { 0 };
+	DatabaseCatalog *sourceDB = specs->sourceDB;
 
-	if (!pgsql_init(&pgsql, specs->connStrings->source_pguri, PGSQL_CONN_SOURCE))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	if (!pgsql_begin(&pgsql))
+	if (!catalog_open(sourceDB))
 	{
 		/* errors have already been logged */
 		return false;
@@ -159,20 +163,14 @@ follow_init_sentinel(StreamSpecs *specs, CopyDBSentinel *sentinel)
 
 	if (specs->endpos != InvalidXLogRecPtr)
 	{
-		if (!pgsql_update_sentinel_endpos(&pgsql, false, specs->endpos))
+		if (!sentinel_update_endpos(sourceDB, specs->endpos))
 		{
 			/* errors have already been logged */
 			return false;
 		}
 	}
 
-	if (!pgsql_get_sentinel(&pgsql, sentinel))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	if (!pgsql_commit(&pgsql))
+	if (!sentinel_get(sourceDB, sentinel))
 	{
 		/* errors have already been logged */
 		return false;
@@ -189,15 +187,9 @@ follow_init_sentinel(StreamSpecs *specs, CopyDBSentinel *sentinel)
 bool
 follow_get_sentinel(StreamSpecs *specs, CopyDBSentinel *sentinel, bool verbose)
 {
-	PGSQL pgsql = { 0 };
+	DatabaseCatalog *sourceDB = specs->sourceDB;
 
-	if (!pgsql_init(&pgsql, specs->connStrings->source_pguri, PGSQL_CONN_SOURCE))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	if (!pgsql_get_sentinel(&pgsql, sentinel))
+	if (!sentinel_get(sourceDB, sentinel))
 	{
 		/* errors have already been logged */
 		return false;
@@ -269,11 +261,9 @@ follow_main_loop(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs)
 		return false;
 	}
 
-	/*
-	 * Read the catalogs from the source table from on-file disk if
-	 * the schema dump has been done already.
-	 */
-	if (copySpecs->dirState.schemaDumpIsDone && !copydb_parse_schema_json_file(copySpecs))
+	DatabaseCatalog *sourceDB = &(copySpecs->catalogs.source);
+
+	if (!catalog_open(sourceDB))
 	{
 		/* errors have already been logged */
 		return false;
@@ -539,6 +529,17 @@ followDB(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs)
 		return false;
 	}
 
+	/*
+	 * Before starting sub-processes, make sure to close our SQLite catalogs.
+	 * We open the SQLite catalogs again before returning from this function
+	 * (if only when reaching the end of it and returning true).
+	 */
+	if (!catalog_close(streamSpecs->sourceDB))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	/*
 	 * Prepare the sub-process communication mechanisms, when needed:
 	 *
@@ -605,12 +606,40 @@ followDB(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs)
 		}
 	}
 
+	/*
+	 * Close pipe ends which follow is not using. Otherwise the processes
+	 * like transform and apply which reads from the pipe during replay
+	 * will never see EOF.
+	 */
+	if (streamSpecs->stdOut)
+	{
+		close_fd_or_exit(streamSpecs->pipe_rt[1]);
+		close_fd_or_exit(streamSpecs->pipe_rt[0]);
+	}
+
+	if (streamSpecs->stdIn)
+	{
+		close_fd_or_exit(streamSpecs->pipe_ta[0]);
+		close_fd_or_exit(streamSpecs->pipe_ta[1]);
+	}
+
 	/*
 	 * Finally wait until the process are finished.
 	 *
 	 * This happens when the sentinel endpos is set, typically using the
 	 * command: pgcopydb stream sentinel set endpos --current.
+	 *
+	 * When waipid() catches a subprocess termination, we need to update our
+	 * sentinel values, and for that we need to catalogs open again. The caller
+	 * to this function had the catalogs open, so we let them opened when
+	 * returning here.
 	 */
+	if (!catalog_open(streamSpecs->sourceDB))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	if (follow_wait_subprocesses(streamSpecs))
 	{
 		log_info("Subprocesses for %s, %s, and %s have now all exited",
@@ -718,6 +747,8 @@ follow_start_transform(StreamSpecs *specs)
 
 		bool success = stream_transform_stream(specs);
 
+		log_info("Transform process has terminated");
+
 		close_fd_or_exit(specs->pipe_rt[0]);
 		close_fd_or_exit(specs->pipe_ta[1]);
 
@@ -734,7 +765,11 @@ follow_start_transform(StreamSpecs *specs)
 		specs->stdIn = false;
 		specs->stdOut = false;
 
-		return stream_transform_worker(specs);
+		bool success = stream_transform_worker(specs);
+
+		log_info("Transform process has terminated");
+
+		return success;
 	}
 
 	return true;
@@ -766,6 +801,8 @@ follow_start_catchup(StreamSpecs *specs)
 
 		bool success = stream_apply_replay(specs);
 
+		log_info("Apply process has terminated");
+
 		close_fd_or_exit(specs->pipe_ta[0]);
 
 		return success;
@@ -781,7 +818,11 @@ follow_start_catchup(StreamSpecs *specs)
 		specs->stdIn = false;
 		specs->stdOut = false;
 
-		return stream_apply_catchup(specs);
+		bool success = stream_apply_catchup(specs);
+
+		log_info("Apply process has terminated");
+
+		return success;
 	}
 
 	return true;
@@ -819,6 +860,7 @@ follow_start_subprocess(StreamSpecs *specs, FollowSubProcess *subprocess)
 		case 0:
 		{
 			/* child process runs the command */
+			pid_t pid = getpid();
 			char psTitle[BUFSIZE] = { 0 };
 
 			sformat(psTitle, sizeof(psTitle), "pgcopydb: follow %s",
@@ -826,6 +868,29 @@ follow_start_subprocess(StreamSpecs *specs, FollowSubProcess *subprocess)
 
 			(void) set_ps_title(psTitle);
 
+			/* also track the process information in our catalogs */
+			ProcessInfo ps = {
+				.pid = pid,
+				.psTitle = ps_buffer
+			};
+
+			strlcpy(ps.psType, subprocess->name, sizeof(ps.psType));
+
+			DatabaseCatalog *sourceDB = specs->sourceDB;
+
+			if (!catalog_open(sourceDB))
+			{
+				/* errors have already been logged */
+				return false;
+			}
+
+			if (!catalog_upsert_process_info(sourceDB, &ps))
+			{
+				log_error("Failed to track progress in our catalogs, "
+						  "see above for details");
+				return false;
+			}
+
 			log_notice("Starting the %s sub-process", subprocess->name);
 
 			if (!(subprocess->command)(specs))
@@ -927,10 +992,29 @@ follow_wait_subprocesses(StreamSpecs *specs)
 			{
 				--stillRunning;
 
+				/*
+				 * First, delete the process from our tracking in the catalogs.
+				 */
+				if (!catalog_delete_process(specs->sourceDB,
+											processArray[i]->pid))
+				{
+					log_notice("Failed to delete process entry for pid %d",
+							   processArray[i]->pid);
+				}
+
 				int logLevel = LOG_NOTICE;
 				char details[BUFSIZE] = { 0 };
 
-				if (processArray[i]->returnCode == 0)
+				/*
+				 * A sub-process exit is considered a "successful" exit when
+				 * the return code is zero and the signal for termination is a
+				 * signal that pgcopydb knows to handle and expects.
+				 */
+				bool exitedSuccessfully =
+					processArray[i]->returnCode == 0 &&
+					signal_is_handled(processArray[i]->sig);
+
+				if (exitedSuccessfully)
 				{
 					if (processArray[i]->sig == 0)
 					{
@@ -947,9 +1031,17 @@ follow_wait_subprocesses(StreamSpecs *specs)
 				{
 					logLevel = LOG_ERROR;
 
-					if (processArray[i]->sig == 0)
+					if (processArray[i]->returnCode == 0)
 					{
-						sformat(details, sizeof(details), "with error code %d",
+						sformat(details, sizeof(details),
+								"with return code %d and signal %s",
+								processArray[i]->returnCode,
+								signal_to_string(processArray[i]->sig));
+					}
+					else if (processArray[i]->sig == 0)
+					{
+						sformat(details, sizeof(details),
+								"with error code %d",
 								processArray[i]->returnCode);
 					}
 					else
@@ -987,7 +1079,7 @@ follow_wait_subprocesses(StreamSpecs *specs)
 					log_warn("Failed to get sentinel values");
 				}
 
-				if (processArray[i]->returnCode != 0 ||
+				if (!exitedSuccessfully ||
 					specs->endpos == InvalidXLogRecPtr)
 				{
 					char endposStatus[BUFSIZE] = { 0 };
@@ -1003,11 +1095,14 @@ follow_wait_subprocesses(StreamSpecs *specs)
 								LSN_FORMAT_ARGS(specs->endpos));
 					}
 
-					log_notice("Process %s has exited with return code %d, "
+					const char *exitmode =
+						exitedSuccessfully ? "successfully" : "unexpectedly";
+
+					log_notice("Process %s has exited %s, "
 							   "and endpos is %s: "
 							   "terminating other processes",
 							   processArray[i]->name,
-							   processArray[i]->returnCode,
+							   exitmode,
 							   endposStatus);
 
 					if (!follow_terminate_subprocesses(specs))
@@ -1018,7 +1113,7 @@ follow_wait_subprocesses(StreamSpecs *specs)
 					}
 				}
 
-				success = success && processArray[i]->returnCode == 0;
+				success = success && exitedSuccessfully;
 			}
 		}
 
