@@ -21,6 +21,7 @@
 #include "copydb.h"
 #include "env_utils.h"
 #include "ld_stream.h"
+#include "lsn_tracking.h"
 #include "lock_utils.h"
 #include "log.h"
 #include "parsing_utils.h"
@@ -45,6 +46,13 @@ GUC applySettings[] = {
 	{ NULL, NULL },
 };
 
+static bool readTxnCommitLSN(LogicalMessageMetadata *metadata, const char *dir,
+							 bool *txnCommitLSNFound);
+static bool parseTxnMetadataFile(const char *filename, LogicalMessageMetadata *metadata);
+
+static bool computeTxnMetadataFilename(uint32_t xid, const char *dir, char *filename);
+
+static bool writeTxnCommitMetadata(LogicalMessageMetadata *mesg, const char *dir);
 
 /*
  * stream_apply_catchup catches up with SQL files that have been prepared by
@@ -174,6 +182,7 @@ stream_apply_setup(StreamSpecs *specs, StreamApplyContext *context)
 {
 	/* init our context */
 	if (!stream_apply_init_context(context,
+								   specs->sourceDB,
 								   &(specs->paths),
 								   specs->connStrings,
 								   specs->origin,
@@ -184,7 +193,7 @@ stream_apply_setup(StreamSpecs *specs, StreamApplyContext *context)
 	}
 
 	/* read-in the previous lsn tracking file, if it exists */
-	if (!stream_apply_read_lsn_tracking(context))
+	if (!lsn_tracking_read(context))
 	{
 		log_error("Failed to read LSN tracking file");
 		return false;
@@ -268,19 +277,9 @@ stream_apply_setup(StreamSpecs *specs, StreamApplyContext *context)
 bool
 stream_apply_wait_for_sentinel(StreamSpecs *specs, StreamApplyContext *context)
 {
-	PGSQL src = { 0 };
 	bool firstLoop = true;
 	CopyDBSentinel sentinel = { 0 };
 
-	if (!pgsql_init(&src, specs->connStrings->source_pguri, PGSQL_CONN_SOURCE))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	/* skip logging the sentinel queries, we log_debug the values fetched */
-	src.logSQL = context->logSQL;
-
 	/* make sure context->apply is false before entering the loop */
 	context->apply = false;
 
@@ -295,7 +294,7 @@ stream_apply_wait_for_sentinel(StreamSpecs *specs, StreamApplyContext *context)
 		}
 
 		/* this reconnects on each loop iteration, every 10s by default */
-		if (!pgsql_get_sentinel(&src, &sentinel))
+		if (!sentinel_get(specs->sourceDB, &sentinel))
 		{
 			log_warn("Retrying to fetch pgcopydb sentinel values in %ds",
 					 CATCHINGUP_SLEEP_MS / 10);
@@ -318,7 +317,7 @@ stream_apply_wait_for_sentinel(StreamSpecs *specs, StreamApplyContext *context)
 		{
 			context->endpos = sentinel.endpos;
 		}
-		else
+		else if (context->endpos != sentinel.endpos)
 		{
 			log_warn("Sentinel endpos is %X/%X, overriden by --endpos %X/%X",
 					 LSN_FORMAT_ARGS(sentinel.endpos),
@@ -377,25 +376,13 @@ stream_apply_wait_for_sentinel(StreamSpecs *specs, StreamApplyContext *context)
 bool
 stream_apply_sync_sentinel(StreamApplyContext *context, bool findDurableLSN)
 {
-	PGSQL src = { 0 };
-	CopyDBSentinel sentinel = { 0 };
-
 	/* now is a good time to write the LSN tracking to disk */
-	if (!stream_apply_write_lsn_tracking(context))
+	if (!lsn_tracking_write(context->sourceDB, context->lsnTrackingList))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	if (!pgsql_init(&src, context->connStrings->source_pguri, PGSQL_CONN_SOURCE))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	/* limit the amount of logging of the apply process */
-	src.logSQL = context->logSQL;
-
 	uint64_t durableLSN = InvalidXLogRecPtr;
 
 	/*
@@ -415,7 +402,9 @@ stream_apply_sync_sentinel(StreamApplyContext *context, bool findDurableLSN)
 		}
 	}
 
-	if (!pgsql_sync_sentinel_apply(&src, durableLSN, &sentinel))
+	CopyDBSentinel sentinel = { 0 };
+
+	if (!sentinel_sync_apply(context->sourceDB, durableLSN, &sentinel))
 	{
 		log_warn("Failed to sync progress with the pgcopydb sentinel");
 		return true;
@@ -425,85 +414,15 @@ stream_apply_sync_sentinel(StreamApplyContext *context, bool findDurableLSN)
 	context->endpos = sentinel.endpos;
 	context->startpos = sentinel.startpos;
 
-	return true;
-}
-
-
-/*
- * stream_apply_send_sync_sentinel sends a query to sync with the pgcopydb
- * sentinel table using the libpq async API. Use the associated function
- * stream_apply_fetch_sync_sentinel to check for availability of the result and
- * fetch it when it's available.
- */
-bool
-stream_apply_send_sync_sentinel(StreamApplyContext *context)
-{
-	PGSQL *src = &(context->src);
-
-	if (context->sentinelQueryInProgress)
-	{
-		log_error("BUG: stream_apply_send_sync_sentinel already in progress");
-		return false;
-	}
-
-	/* we're going to keep the connection around */
-	context->src.connectionStatementType = PGSQL_CONNECTION_MULTI_STATEMENT;
-
-	/* limit the amount of logging of the apply process */
-	src->logSQL = true;
-
-	uint64_t durableLSN = InvalidXLogRecPtr;
-
-	if (!stream_apply_find_durable_lsn(context, &durableLSN))
-	{
-		log_debug("Skipping sentinel replay_lsn update, "
-				  "failed to find a durable LSN matching current flushLSN");
-		return true;
-	}
-
-	if (!pgsql_send_sync_sentinel_apply(src, durableLSN))
-	{
-		log_warn("Failed to sync progress with the pgcopydb sentinel");
-		return true;
-	}
-
-	context->sentinelSyncTime = time(NULL);
-	context->sentinelQueryInProgress = true;
-
-	return true;
-}
-
-
-/*
- * stream_apply_fetch_sync_sentinel checks to see if the result of the
- * pgcopydb sentinel sync query is available and fetches it then.
- */
-bool
-stream_apply_fetch_sync_sentinel(StreamApplyContext *context)
-{
-	PGSQL *src = &(context->src);
-
-	bool retry;
-	CopyDBSentinel sentinel = { 0 };
-
-	if (!pgsql_fetch_sync_sentinel_apply(src, &retry, &sentinel))
-	{
-		log_error("Failed to fetch sentinel update query results");
-		return false;
-	}
-
-	if (!retry)
-	{
-		context->sentinelQueryInProgress = false;
-
-		/* also disconnect between async queries */
-		(void) pgsql_finish(&(context->src));
-
-		context->apply = sentinel.apply;
-		context->endpos = sentinel.endpos;
-		context->startpos = sentinel.startpos;
-		context->replay_lsn = sentinel.replay_lsn;
-	}
+	log_debug("stream_apply_sync_sentinel: "
+			  "write_lsn %X/%X flush_lsn %X/%X replay_lsn %X/%X "
+			  "startpos %X/%X endpos %X/%X apply %s",
+			  LSN_FORMAT_ARGS(sentinel.write_lsn),
+			  LSN_FORMAT_ARGS(sentinel.flush_lsn),
+			  LSN_FORMAT_ARGS(sentinel.replay_lsn),
+			  LSN_FORMAT_ARGS(context->startpos),
+			  LSN_FORMAT_ARGS(context->endpos),
+			  context->apply ? "enabled" : "disabled");
 
 	return true;
 }
@@ -671,6 +590,16 @@ stream_apply_sql(StreamApplyContext *context,
 				return false;
 			}
 
+			bool txnCommitLSNFound = false;
+
+			if (!readTxnCommitLSN(metadata,
+								  context->paths.dir,
+								  &txnCommitLSNFound))
+			{
+				log_error("Failed to read transaction metadata file");
+				return false;
+			}
+
 			/*
 			 * Few a time, BEGIN won't have a txnCommitLSN for the txn which
 			 * spread across multiple WAL segments. We call that txn as
@@ -680,7 +609,7 @@ stream_apply_sql(StreamApplyContext *context,
 			 * The lsn of a COMMIT message determines whether to keep txn or
 			 * abort.
 			 */
-			context->continuedTxn = metadata->txnCommitLSN == InvalidXLogRecPtr;
+			context->continuedTxn = !txnCommitLSNFound;
 
 			/* did we reach the starting LSN positions now? */
 			if (!context->reachedStartPos)
@@ -783,10 +712,47 @@ stream_apply_sql(StreamApplyContext *context,
 			break;
 		}
 
+		case STREAM_ACTION_ROLLBACK:
+		{
+			/* Rollback the transaction */
+			if (!pgsql_execute(pgsql, "ROLLBACK"))
+			{
+				/* errors have already been logged */
+				return false;
+			}
+
+			/* Reset the transactionInProgress after abort */
+			context->transactionInProgress = false;
+
+			/* Reevaluate reachedStartPos after rollback */
+			context->reachedStartPos = false;
+
+			break;
+		}
+
 		case STREAM_ACTION_COMMIT:
 		{
 			context->reachedStartPos = context->previousLSN < metadata->lsn;
 
+			if (context->continuedTxn)
+			{
+				/*
+				 * Write the transaction metadata file for continuedTxn.
+				 * This file will be used for the resumed transaction
+				 * to determine whether allow the transaction to be
+				 * replayed or not.
+				 * Without this, executing the same continuedTxn twice
+				 * will result in duplicate key errors if the table has
+				 * unique constraints.
+				 */
+				if (!writeTxnCommitMetadata(metadata, context->paths.dir))
+				{
+					log_error("Failed to write transaction metadata file, "
+							  "see above for details");
+					return false;
+				}
+			}
+
 			if (!context->reachedStartPos)
 			{
 				/*
@@ -892,18 +858,6 @@ stream_apply_sql(StreamApplyContext *context,
 					  LSN_FORMAT_ARGS(metadata->lsn),
 					  LSN_FORMAT_ARGS(context->previousLSN));
 
-			/*
-			 * Don't update previousLSN if we are in a continuedTxn.
-			 *
-			 * Otherwise, during resume a continued txn having an endpos
-			 * will update the previousLSN to endpos's LSN causing the
-			 * current transaction to be applied again.
-			 */
-			if (!context->continuedTxn)
-			{
-				context->previousLSN = metadata->lsn;
-			}
-
 			/*
 			 * It could be the current endpos, or the endpos of a previous
 			 * run.
@@ -911,6 +865,7 @@ stream_apply_sql(StreamApplyContext *context,
 			if (context->endpos != InvalidXLogRecPtr &&
 				context->endpos <= metadata->lsn)
 			{
+				context->previousLSN = metadata->lsn;
 				context->reachedEndPos = true;
 
 				log_notice("Apply reached end position %X/%X at ENDPOS %X/%X",
@@ -1083,9 +1038,7 @@ stream_apply_sql(StreamApplyContext *context,
 					return false;
 				}
 
-				PreparedStmt *stmt =
-					(PreparedStmt *) calloc(1, sizeof(PreparedStmt));
-
+				stmt = (PreparedStmt *) calloc(1, sizeof(PreparedStmt));
 				stmt->hash = hash;
 				stmt->prepared = true;
 
@@ -1234,6 +1187,17 @@ setupReplicationOrigin(StreamApplyContext *context, bool logSQL)
 	/* we also might want to skip logging any SQL query that we apply */
 	pgsql->logSQL = logSQL;
 
+	/*
+	 * Grab the Postgres server version on the target, we need to know that for
+	 * being able to call pgsql_current_wal_insert_lsn using the right Postgres
+	 * function name.
+	 */
+	if (!pgsql_server_version(pgsql))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	uint32_t oid = 0;
 
 	if (!pgsql_replication_origin_oid(pgsql, nodeName, &oid))
@@ -1324,11 +1288,13 @@ setupReplicationOrigin(StreamApplyContext *context, bool logSQL)
  */
 bool
 stream_apply_init_context(StreamApplyContext *context,
+						  DatabaseCatalog *sourceDB,
 						  CDCPaths *paths,
 						  ConnStrings *connStrings,
 						  char *origin,
 						  uint64_t endpos)
 {
+	context->sourceDB = sourceDB;
 	context->paths = *paths;
 
 	/*
@@ -1425,6 +1391,7 @@ parseSQLAction(const char *query, LogicalMessageMetadata *metadata)
 	char *message = NULL;
 	char *begin = strstr(query, OUTPUT_BEGIN);
 	char *commit = strstr(query, OUTPUT_COMMIT);
+	char *rollback = strstr(query, OUTPUT_ROLLBACK);
 	char *switchwal = strstr(query, OUTPUT_SWITCHWAL);
 	char *keepalive = strstr(query, OUTPUT_KEEPALIVE);
 	char *endpos = strstr(query, OUTPUT_ENDPOS);
@@ -1440,6 +1407,11 @@ parseSQLAction(const char *query, LogicalMessageMetadata *metadata)
 		metadata->action = STREAM_ACTION_COMMIT;
 		message = commit + strlen(OUTPUT_COMMIT);
 	}
+	else if (query == rollback)
+	{
+		metadata->action = STREAM_ACTION_ROLLBACK;
+		message = rollback + strlen(OUTPUT_ROLLBACK);
+	}
 	else if (query == switchwal)
 	{
 		metadata->action = STREAM_ACTION_SWITCH;
@@ -1630,13 +1602,14 @@ stream_apply_track_insert_lsn(StreamApplyContext *context, uint64_t sourceLSN)
 bool
 stream_apply_find_durable_lsn(StreamApplyContext *context, uint64_t *durableLSN)
 {
-	PGSQL *pgsql = &(context->pgsql);
-
 	uint64_t flushLSN = InvalidXLogRecPtr;
 
-	if (!pgsql_current_wal_flush_lsn(pgsql, &flushLSN))
+	if (!stream_fetch_current_lsn(&flushLSN,
+								  context->connStrings->target_pguri,
+								  PGSQL_CONN_SOURCE))
 	{
-		/* errors have already been logged */
+		log_error("Failed to retrieve current WAL positions, "
+				  "see above for details");
 		return false;
 	}
 
@@ -1684,128 +1657,175 @@ stream_apply_find_durable_lsn(StreamApplyContext *context, uint64_t *durableLSN)
 
 
 /*
- * stream_apply_write_lsn_tracking writes the context->LSNTracking linked-list
- * as a JSON array on-disk.
+ * readTxnCommitLSN ensures metadata has transaction COMMIT LSN by fetching it
+ * from metadata file if it is not present
  */
-bool
-stream_apply_write_lsn_tracking(StreamApplyContext *context)
+static bool
+readTxnCommitLSN(LogicalMessageMetadata *metadata,
+				 const char *dir,
+				 bool *txnCommitLSNFound)
 {
-	const char *filename = context->paths.lsntrackingfile;
-	LSNTracking *current = context->lsnTrackingList;
-
-	JSON_Value *js = json_value_init_array();
-	JSON_Array *jsArray = json_value_get_array(js);
-
-	for (; current != NULL; current = current->previous)
+	/* if txnCommitLSN is invalid, then fetch it from txn metadata file */
+	if (metadata->txnCommitLSN != InvalidXLogRecPtr)
 	{
-		JSON_Value *jsObj = json_value_init_object();
-		JSON_Object *jsLSNObj = json_value_get_object(jsObj);
-
-		char sourceLSN[PG_LSN_MAXLENGTH] = { 0 };
-		char insertLSN[PG_LSN_MAXLENGTH] = { 0 };
-
-		sformat(sourceLSN, sizeof(sourceLSN), "%X/%X",
-				LSN_FORMAT_ARGS(current->sourceLSN));
-
-		sformat(insertLSN, sizeof(insertLSN), "%X/%X",
-				LSN_FORMAT_ARGS(current->insertLSN));
-
-		json_object_set_string(jsLSNObj, "source", sourceLSN);
-		json_object_set_string(jsLSNObj, "insert", insertLSN);
-
-		json_array_append_value(jsArray, jsObj);
+		*txnCommitLSNFound = true;
+		return true;
 	}
 
-	char *serialized_string = json_serialize_to_string(js);
-	size_t len = strlen(serialized_string);
+	char txnfilename[MAXPGPATH] = { 0 };
 
-	if (!write_file(serialized_string, len, filename))
+	if (!computeTxnMetadataFilename(metadata->xid,
+									dir,
+									txnfilename))
 	{
-		log_error("Failed to write LSN tracking to file \"%s\"", filename);
+		/* errors have already been logged */
 		return false;
 	}
 
-	json_free_serialized_string(serialized_string);
-	json_value_free(js);
+	if (!file_exists(txnfilename))
+	{
+		*txnCommitLSNFound = false;
+		return true;
+	}
+
+	log_debug("stream_apply_sql: BEGIN message without a commit LSN, "
+			  "fetching commit LSN from transaction metadata file \"%s\"",
+			  txnfilename);
+
+	LogicalMessageMetadata txnMetadata = { .xid = metadata->xid };
+
+	if (!parseTxnMetadataFile(txnfilename, &txnMetadata))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	*txnCommitLSNFound = true;
+	metadata->txnCommitLSN = txnMetadata.txnCommitLSN;
 
 	return true;
 }
 
 
 /*
- * stream_apply_read_lsn_tracking reads the context->LSNTracking linked-list
- * from a JSON array on-disk.
+ * parseTxnMetadataFile returns the transaction metadata content for the given
+ * metadata filename.
  */
-bool
-stream_apply_read_lsn_tracking(StreamApplyContext *context)
+static bool
+parseTxnMetadataFile(const char *filename, LogicalMessageMetadata *metadata)
 {
-	const char *filename = context->paths.lsntrackingfile;
-	LSNTracking *current = context->lsnTrackingList;
+	/* store xid as it will be overwritten while parsing metadata */
+	uint32_t xid = metadata->xid;
 
-	if (current != NULL)
+	if (xid == 0)
 	{
-		log_error("BUG: stream_apply_read_lsn_tracking current is not NULL");
+		log_error("BUG: parseTxnMetadataFile is called with "
+				  "transaction xid: %lld", (long long) xid);
 		return false;
 	}
 
-	/* it's okay if the file does not exists, just skip the operation */
-	if (!file_exists(filename))
-	{
-		log_notice("Failed to parse JSON file \"%s\": file does not exists",
-				   filename);
-		return true;
-	}
+	char *txnMetadataContent = NULL;
+	long size = 0L;
 
-	JSON_Value *json = json_parse_file(filename);
-
-	if (json == NULL)
+	if (!read_file(filename, &txnMetadataContent, &size))
 	{
-		log_error("Failed to parse JSON file \"%s\"", filename);
+		/* errors have already been logged */
 		return false;
 	}
 
-	JSON_Array *jsArray = json_value_get_array(json);
-	int count = json_array_get_count(jsArray);
+	JSON_Value *json = json_parse_string(txnMetadataContent);
 
-	for (int i = 0; i < count; i++)
+	if (!parseMessageMetadata(metadata, txnMetadataContent, json, true))
 	{
-		LSNTracking *lsn_tracking =
-			(LSNTracking *) calloc(1, sizeof(LSNTracking));
-
-		if (lsn_tracking == NULL)
-		{
-			log_error(ALLOCATION_FAILED_ERROR);
-
-			(void) json_value_free(json);
-			return false;
-		}
-
-		JSON_Object *jsObj = json_array_get_object(jsArray, i);
-		const char *source = json_object_get_string(jsObj, "source");
-		const char *insert = json_object_get_string(jsObj, "insert");
-
-		if (!parseLSN(source, &(lsn_tracking->sourceLSN)))
-		{
-			log_error("Failed to parse source LSN \"%s\"", source);
-
-			(void) json_value_free(json);
-			return false;
-		}
-
-		if (!parseLSN(insert, &(lsn_tracking->insertLSN)))
-		{
-			log_error("Failed to parse insert LSN \"%s\"", insert);
-
-			(void) json_value_free(json);
-			return false;
-		}
-
-		/* update the linked list */
-		lsn_tracking->previous = context->lsnTrackingList;
-		context->lsnTrackingList = lsn_tracking;
+		/* errors have already been logged */
+		json_value_free(json);
+		return false;
 	}
 
-	(void) json_value_free(json);
+	json_value_free(json);
+	free(txnMetadataContent);
+
+	if (metadata->txnCommitLSN == InvalidXLogRecPtr ||
+		metadata->xid != xid ||
+		IS_EMPTY_STRING_BUFFER(metadata->timestamp))
+	{
+		log_error("Failed to parse metadata for transaction metadata file "
+				  "\"%s\": %s", filename, txnMetadataContent);
+		return false;
+	}
+
+	return true;
+}
+
+
+/*
+ *  computeTxnMetadataFilename computes the file path for transaction metadata
+ *  based on its transaction id
+ */
+static bool
+computeTxnMetadataFilename(uint32_t xid, const char *dir, char *filename)
+{
+	if (dir == NULL)
+	{
+		log_error("BUG: computeTxnMetadataFilename is called with "
+				  "directory: NULL");
+		return false;
+	}
+
+	if (xid == 0)
+	{
+		log_error("BUG: computeTxnMetadataFilename is called with "
+				  "transaction xid: %lld", (long long) xid);
+		return false;
+	}
+
+	sformat(filename, MAXPGPATH, "%s/%lld.json", dir, (long long) xid);
+
+	return true;
+}
+
+
+/*
+ * writeTxnCommitMetadata writes the transaction metadata to a file in the given
+ * directory
+ */
+static bool
+writeTxnCommitMetadata(LogicalMessageMetadata *mesg, const char *dir)
+{
+	char txnfilename[MAXPGPATH] = { 0 };
+
+	if (mesg->action != STREAM_ACTION_COMMIT)
+	{
+		log_error("BUG: writeTxnCommitMetadata is called with "
+				  "action: %s", StreamActionToString(mesg->action));
+		return false;
+	}
+
+	if (!computeTxnMetadataFilename(mesg->xid, dir, txnfilename))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	log_debug("stream_write_commit_metadata_file: writing transaction "
+			  "metadata file \"%s\" with commit lsn %X/%X",
+			  txnfilename,
+			  LSN_FORMAT_ARGS(mesg->lsn));
+
+	char contents[BUFSIZE] = { 0 };
+
+	sformat(contents, BUFSIZE,
+			"{\"xid\":%lld,\"commit_lsn\":\"%X/%X\",\"timestamp\":\"%s\"}\n",
+			(long long) mesg->xid,
+			LSN_FORMAT_ARGS(mesg->lsn),
+			mesg->timestamp);
+
+	/* write the metadata to txnfilename */
+	if (!write_file(contents, strlen(contents), txnfilename))
+	{
+		log_error("Failed to write file \"%s\"", txnfilename);
+		return false;
+	}
 
 	return true;
 }
