@@ -48,7 +48,7 @@ stream_init_specs(StreamSpecs *specs,
 				  char *origin,
 				  uint64_t endpos,
 				  LogicalStreamMode mode,
-				  SourceCatalog *catalog,
+				  DatabaseCatalog *sourceDB,
 				  bool stdin,
 				  bool stdout,
 				  bool logSQL)
@@ -62,7 +62,16 @@ stream_init_specs(StreamSpecs *specs,
 	specs->paths = *paths;
 	specs->endpos = endpos;
 
-	specs->catalog = catalog;
+	/*
+	 * Open the specified sourceDB catalog.
+	 */
+	specs->sourceDB = sourceDB;
+
+	if (!catalog_init(specs->sourceDB))
+	{
+		/* errors have already been logged */
+		return false;
+	}
 
 	/*
 	 * Copy the given ReplicationSlot: it comes from command line parsing, or
@@ -92,14 +101,16 @@ stream_init_specs(StreamSpecs *specs,
 		case STREAM_PLUGIN_WAL2JSON:
 		{
 			KeyVal options = {
-				.count = 6,
+				/* we ignore the last keyword and value if the option is not set */
+				.count = specs->slot.wal2jsonNumericAsString ? 7 : 6,
 				.keywords = {
 					"format-version",
 					"include-xids",
 					"include-schemas",
 					"include-transaction",
 					"include-types",
-					"filter-tables"
+					"filter-tables",
+					"numeric-data-types-as-string"
 				},
 				.values = {
 					"2",
@@ -107,12 +118,12 @@ stream_init_specs(StreamSpecs *specs,
 					"true",
 					"true",
 					"true",
-					"pgcopydb.*"
+					"pgcopydb.*",
+					"true"
 				}
 			};
 
 			specs->pluginOptions = options;
-
 			break;
 		}
 
@@ -338,14 +349,12 @@ stream_init_context(StreamSpecs *specs)
 
 	privateContext->endpos = specs->endpos;
 	privateContext->startpos = specs->startpos;
-	privateContext->startposActionFromJSON = specs->startposActionFromJSON;
 
 	privateContext->mode = specs->mode;
 
 	privateContext->transformQueue = &(specs->transformQueue);
 
 	privateContext->paths = specs->paths;
-	privateContext->startpos = specs->startpos;
 
 	privateContext->connStrings = specs->connStrings;
 
@@ -379,20 +388,6 @@ stream_init_context(StreamSpecs *specs)
 
 	privateContext->lastWriteTime = 0;
 
-	if (specs->startposComputedFromJSON &&
-		(specs->startposActionFromJSON == STREAM_ACTION_BEGIN ||
-		 specs->startposActionFromJSON == STREAM_ACTION_INSERT ||
-		 specs->startposActionFromJSON == STREAM_ACTION_UPDATE ||
-		 specs->startposActionFromJSON == STREAM_ACTION_DELETE ||
-		 specs->startposActionFromJSON == STREAM_ACTION_TRUNCATE))
-	{
-		privateContext->reachedStartPos = false;
-	}
-	else
-	{
-		privateContext->reachedStartPos = true;
-	}
-
 	/*
 	 * Initializing maxWrittenLSN as startpos at the beginning of migration or
 	 * when resuming from interruption where it will be equal to
@@ -413,7 +408,7 @@ stream_init_context(StreamSpecs *specs)
 	privateContext->maxWrittenLSN = specs->startpos;
 
 	/* transform needs some catalog lookups (pkey, type oid) */
-	privateContext->catalog = specs->catalog;
+	privateContext->sourceDB = specs->sourceDB;
 
 	return true;
 }
@@ -591,17 +586,9 @@ streamCheckResumePosition(StreamSpecs *specs)
 	 * command line option (found in specs->endpos) prevails, but when it's not
 	 * been used, we have a look at the sentinel value.
 	 */
-	PGSQL src = { 0 };
-
-	if (!pgsql_init(&src, specs->connStrings->source_pguri, PGSQL_CONN_SOURCE))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
 	CopyDBSentinel sentinel = { 0 };
 
-	if (!pgsql_get_sentinel(&src, &sentinel))
+	if (!sentinel_get(specs->sourceDB, &sentinel))
 	{
 		/* errors have already been logged */
 		return false;
@@ -622,7 +609,7 @@ streamCheckResumePosition(StreamSpecs *specs)
 					 LSN_FORMAT_ARGS(specs->endpos));
 		}
 
-		if (!pgsql_update_sentinel_endpos(&src, false, specs->endpos))
+		if (!sentinel_update_endpos(specs->sourceDB, specs->endpos))
 		{
 			/* errors have already been logged */
 			return false;
@@ -660,48 +647,25 @@ streamCheckResumePosition(StreamSpecs *specs)
 		LogicalMessageMetadata *messages = latestStreamedContent.messages;
 		LogicalMessageMetadata *latest = &(messages[lastLineNb]);
 
-		/*
-		 * We could have several messages following each-other with the same
-		 * LSN, typically a sequence like:
-		 *
-		 *  {"action":"I","xid":"492","lsn":"0/244BEE0", ...}
-		 *  {"action":"K","lsn":"0/244BEE0", ...}
-		 *  {"action":"E","lsn":"0/244BEE0"}
-		 *
-		 * In that case we want to remember the latest message action as being
-		 * INSERT rather than ENDPOS.
-		 */
-		int lineNb = lastLineNb;
-
-		for (; lineNb > 0; lineNb--)
-		{
-			LogicalMessageMetadata *previous = &(messages[lineNb]);
-
-			if (previous->lsn == latest->lsn)
-			{
-				latest = previous;
-			}
-			else
-			{
-				break;
-			}
-		}
-
 		specs->startpos = latest->lsn;
-		specs->startposComputedFromJSON = true;
-		specs->startposActionFromJSON = latest->action;
 
 		log_info("Resuming streaming at LSN %X/%X "
-				 "from first message with that LSN read in JSON file \"%s\", "
-				 "line %d",
+				 "from JSON file \"%s\" ",
 				 LSN_FORMAT_ARGS(specs->startpos),
-				 latestStreamedContent.filename,
-				 lineNb);
+				 latestStreamedContent.filename);
 
-		char *latestMessage = latestStreamedContent.lines[lineNb];
+		char *latestMessage = latestStreamedContent.lines[lastLineNb];
 		log_notice("Resume replication from latest message: %s", latestMessage);
 	}
 
+	PGSQL src = { 0 };
+
+	if (!pgsql_init(&src, specs->connStrings->source_pguri, PGSQL_CONN_SOURCE))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	bool flush = false;
 	uint64_t lsn = 0;
 
@@ -899,6 +863,30 @@ stream_write_json(LogicalStreamContext *context, bool previous)
 	destroyPQExpBuffer(buffer);
 	free(metadata->jsonBuffer);
 
+	/*
+	 * Maintain the transaction progress based on the BEGIN and COMMIT messages
+	 * received from replication slot. We don't care about the other messages.
+	 */
+	if (metadata->action == STREAM_ACTION_BEGIN)
+	{
+		privateContext->transactionInProgress = true;
+	}
+	else if (metadata->action == STREAM_ACTION_COMMIT)
+	{
+		privateContext->transactionInProgress = false;
+	}
+
+	/*
+	 * We are not expecting STREAM_ACTION_ROLLBACK here. It's a custom
+	 * message we write directly to the "latest" file using
+	 * stream_write_internal_message to abort the last incomplete transaction.
+	 */
+	else if (metadata->action == STREAM_ACTION_ROLLBACK)
+	{
+		log_error("BUG: STREAM_ACTION_ROLLBACK is not expected here");
+		return false;
+	}
+
 	return true;
 }
 
@@ -1237,6 +1225,33 @@ streamCloseFile(LogicalStreamContext *context, bool time_to_abort)
 		}
 	}
 
+	/*
+	 * On graceful exit, ROLLBACK the last incomplete transaction.
+	 * As we resume from a consistent point, there's no concern about
+	 * the transaction being rolled back here.
+	 *
+	 * TODO: For process crashes (e.g., segmentation faults), this
+	 * method won't work, potentially leaving incomplete transactions.
+	 * To handle this, we should read the last message from the "latest"
+	 * file and rollback any incomplete transaction found.
+	 */
+	if (time_to_abort &&
+		privateContext->jsonFile != NULL &&
+		privateContext->transactionInProgress)
+	{
+		InternalMessage rollback = {
+			.action = STREAM_ACTION_ROLLBACK,
+			.lsn = context->cur_record_lsn
+		};
+
+		if (!stream_write_internal_message(context, &rollback))
+		{
+			/* errors have already been logged */
+			return false;
+		}
+	}
+
+
 	/*
 	 * If we have a JSON file currently opened, then close it.
 	 *
@@ -1402,6 +1417,12 @@ streamFlush(LogicalStreamContext *context)
 				  privateContext->partialFileName);
 	}
 
+	/* at flush time also update our internal sentinel tracking */
+	if (!stream_sync_sentinel(context))
+	{
+		/* errors have already been logged */
+		return false;
+	}
 	return true;
 }
 
@@ -1497,14 +1518,12 @@ streamClose(LogicalStreamContext *context)
  *
  * This function is called when it's time to send feedback to the source
  * Postgres instance, include write_lsn, flush_lsn, and replay_lsn. Once in a
- * while we fetch the replay_lsn from the pgcopydb sentinel table on the source
- * database, and sync with the current progress.
+ * while we fetch the replay_lsn from the pgcopydb sentinel table and sync with
+ * the current progress.
  */
 bool
 streamFeedback(LogicalStreamContext *context)
 {
-	StreamContext *privateContext = (StreamContext *) context->private;
-
 	int feedbackInterval = 1 * 1000; /* 1s */
 
 	if (!context->forceFeedback)
@@ -1517,23 +1536,36 @@ streamFeedback(LogicalStreamContext *context)
 		}
 	}
 
-	PGSQL src = { 0 };
-	char *pguri = privateContext->connStrings->source_pguri;
-
-	if (!pgsql_init(&src, pguri, PGSQL_CONN_SOURCE))
+	if (!stream_sync_sentinel(context))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
+	/* mark that we just did a feedback sync */
+	context->lastFeedbackSync = context->now;
+
+	return true;
+}
+
+
+/*
+ * stream_sync_sentinel syncs the sentinel values in our internal catalogs with
+ * the current streaming protocol values.
+ */
+bool
+stream_sync_sentinel(LogicalStreamContext *context)
+{
+	StreamContext *privateContext = (StreamContext *) context->private;
 	CopyDBSentinel sentinel = { 0 };
 
-	if (!pgsql_sync_sentinel_recv(&src,
-								  context->tracking->written_lsn,
-								  context->tracking->flushed_lsn,
-								  &sentinel))
+	if (!sentinel_sync_recv(privateContext->sourceDB,
+							context->tracking->written_lsn,
+							context->tracking->flushed_lsn,
+							&sentinel))
 	{
-		/* errors have already been logged */
+		log_error("Failed to update sentinel at stream flush time, "
+				  "see above for details");
 		return false;
 	}
 
@@ -1548,10 +1580,9 @@ streamFeedback(LogicalStreamContext *context)
 	context->endpos = sentinel.endpos;
 	context->tracking->applied_lsn = sentinel.replay_lsn;
 
-	context->lastFeedbackSync = context->now;
-
-	log_debug("streamFeedback: written %X/%X flushed %X/%X applied %X/%X "
-			  " startpos %X/%X endpos %X/%X apply %s",
+	log_debug("stream_sync_sentinel: "
+			  "write_lsn %X/%X flush_lsn %X/%X apply_lsn %X/%X "
+			  "startpos %X/%X endpos %X/%X apply %s",
 			  LSN_FORMAT_ARGS(context->tracking->written_lsn),
 			  LSN_FORMAT_ARGS(context->tracking->flushed_lsn),
 			  LSN_FORMAT_ARGS(context->tracking->applied_lsn),
@@ -1609,40 +1640,6 @@ prepareMessageMetadataFromContext(LogicalStreamContext *context)
 		return true;
 	}
 
-	/*
-	 * When streaming resumed for a partially applied txn, have we reached a
-	 * message that wasn't flushed in the previous command?
-	 */
-	if (!privateContext->reachedStartPos)
-	{
-		/*
-		 * Also the same LSN might be assigned to a BEGIN message, a COMMIT
-		 * message, and a KEEPALIVE message. Avoid skipping what looks like the
-		 * same message as the latest flushed in our JSON file when it's
-		 * actually a new message.
-		 */
-		privateContext->reachedStartPos =
-			privateContext->startpos < metadata->lsn ||
-			(privateContext->startpos == metadata->lsn &&
-			 metadata->action != privateContext->startposActionFromJSON);
-	}
-
-	if (!privateContext->reachedStartPos)
-	{
-		metadata->filterOut = true;
-
-		log_debug("Skipping write for action %c for XID %u at LSN %X/%X: "
-				  "startpos %X/%X not been reached",
-				  metadata->action,
-				  metadata->xid,
-				  LSN_FORMAT_ARGS(metadata->lsn),
-				  LSN_FORMAT_ARGS(privateContext->startpos));
-
-		*previous = *metadata;
-
-		return true;
-	}
-
 	if (!prepareMessageJSONbuffer(context))
 	{
 		log_error("Failed to prepare a JSON buffer from "
@@ -1911,8 +1908,8 @@ parseMessageMetadata(LogicalMessageMetadata *metadata,
 			if (strlcpy(metadata->timestamp, timestamp, n) >= n)
 			{
 				log_error("Failed to parse JSON message timestamp value \"%s\" "
-						  "which is %lu bytes long, "
-						  "pgcopydb only support timestamps up to %lu bytes",
+						  "which is %zu bytes long, "
+						  "pgcopydb only support timestamps up to %zu bytes",
 						  timestamp,
 						  strlen(timestamp),
 						  sizeof(metadata->timestamp));
@@ -2206,6 +2203,11 @@ StreamActionFromChar(char action)
 			return STREAM_ACTION_ENDPOS;
 		}
 
+		case 'R':
+		{
+			return STREAM_ACTION_ROLLBACK;
+		}
+
 		default:
 		{
 			log_error("Failed to parse JSON message action: \"%c\"", action);
@@ -2281,6 +2283,11 @@ StreamActionToString(StreamAction action)
 			return "ENDPOS";
 		}
 
+		case STREAM_ACTION_ROLLBACK:
+		{
+			return "ROLLBACK";
+		}
+
 		default:
 		{
 			log_error("Failed to parse message action: \"%c\"", action);
@@ -2513,79 +2520,68 @@ stream_create_sentinel(CopyDataSpec *copySpecs,
 		return true;
 	}
 
-	char *sql[] = {
-		"create schema if not exists pgcopydb",
-		"drop table if exists pgcopydb.sentinel",
-		"create table pgcopydb.sentinel"
-		"(startpos pg_lsn, endpos pg_lsn, apply bool, "
-		" write_lsn pg_lsn, flush_lsn pg_lsn, replay_lsn pg_lsn)",
-		NULL
-	};
+	DatabaseCatalog *sourceDB = &(copySpecs->catalogs.source);
 
-	char *index = "create unique index on pgcopydb.sentinel((1))";
+	if (!sentinel_setup(sourceDB, startpos, endpos))
+	{
+		log_error("Failed to create the sentinel table, see above for details");
+		return false;
+	}
 
-	PGSQL *pgsql = &(copySpecs->sourceSnapshot.pgsql);
+	return true;
+}
 
-	if (!pgsql_init(pgsql, copySpecs->connStrings.source_pguri, PGSQL_CONN_SOURCE))
+
+/*
+ * stream_fetch_current_source_lsn connects to the given Postgres service and
+ * fetches the current WAL LSN position by calling pg_current_wal_flush_lsn
+ * there, or the variant of that function that is supported by this Postgres
+ * version.
+ */
+bool
+stream_fetch_current_lsn(uint64_t *lsn,
+						 const char *pguri,
+						 ConnectionType connectionType)
+{
+	PGSQL src = { 0 };
+
+	if (!pgsql_init(&src, (char *) pguri, connectionType))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	if (!pgsql_begin(pgsql))
+	/* limit the amount of logging of the apply process */
+	src.logSQL = false;
+
+	uint64_t flushLSN = InvalidXLogRecPtr;
+
+	if (!pgsql_begin(&src))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	/* create the schema and the table for pgcopydb.sentinel */
-	for (int i = 0; sql[i] != NULL; i++)
-	{
-		log_info("%s", sql[i]);
-
-		if (!pgsql_execute(pgsql, sql[i]))
-		{
-			/* errors have already been logged */
-			return false;
-		}
-	}
-
-	/* now insert the sentinel values (startpos, endpos, false as apply) */
-	char *insert =
-		"insert into pgcopydb.sentinel "
-		"(startpos, endpos, apply, write_lsn, flush_lsn, replay_lsn) "
-		"values($1, $2, $3, '0/0', '0/0', '0/0')";
-
-	char startLSN[PG_LSN_MAXLENGTH] = { 0 };
-	char endLSN[PG_LSN_MAXLENGTH] = { 0 };
-
-	sformat(startLSN, sizeof(startLSN), "%X/%X", LSN_FORMAT_ARGS(startpos));
-	sformat(endLSN, sizeof(endLSN), "%X/%X", LSN_FORMAT_ARGS(endpos));
-
-	int paramCount = 3;
-	Oid paramTypes[3] = { LSNOID, LSNOID, BOOLOID };
-	const char *paramValues[3] = { startLSN, endLSN, "false" };
-
-	if (!pgsql_execute_with_params(pgsql, insert,
-								   paramCount, paramTypes, paramValues,
-								   NULL, NULL))
+	if (!pgsql_server_version(&src))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	if (!pgsql_execute(pgsql, index))
+	if (!pgsql_current_wal_flush_lsn(&src, &flushLSN))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	if (!pgsql_commit(pgsql))
+	if (!pgsql_commit(&src))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
+	*lsn = flushLSN;
+
 	return true;
 }
 
@@ -2657,6 +2653,9 @@ stream_cleanup_context(StreamSpecs *specs)
 	success = success && unlink_file(specs->paths.tlifile);
 	success = success && unlink_file(specs->paths.tlihistfile);
 
+	/* reset the timeline, so that we always read from the disk */
+	specs->system.timeline = 0;
+
 	return success;
 }
 
