@@ -17,6 +17,7 @@
 
 #define OUTPUT_BEGIN "BEGIN; -- "
 #define OUTPUT_COMMIT "COMMIT; -- "
+#define OUTPUT_ROLLBACK "ROLLBACK; -- "
 #define OUTPUT_SWITCHWAL "-- SWITCH WAL "
 #define OUTPUT_KEEPALIVE "-- KEEPALIVE "
 #define OUTPUT_ENDPOS "-- ENDPOS "
@@ -42,7 +43,8 @@ typedef enum
 	STREAM_ACTION_MESSAGE = 'M',
 	STREAM_ACTION_SWITCH = 'X',
 	STREAM_ACTION_KEEPALIVE = 'K',
-	STREAM_ACTION_ENDPOS = 'E'
+	STREAM_ACTION_ENDPOS = 'E',
+	STREAM_ACTION_ROLLBACK = 'R'
 } StreamAction;
 
 typedef struct InternalMessage
@@ -133,32 +135,35 @@ typedef struct LogicalMessageTupleArray
 	LogicalMessageTuple *array; /* malloc'ed area */
 } LogicalMessageTupleArray;
 
+typedef struct LogicalMessageRelation
+{
+	char *nspname;  /* malloc'ed area */
+	char *relname;  /* malloc'ed area */
+	bool pqMemory;
+} LogicalMessageRelation;
+
 typedef struct LogicalMessageInsert
 {
-	char nspname[PG_NAMEDATALEN];
-	char relname[PG_NAMEDATALEN];
+	LogicalMessageRelation table;
 	LogicalMessageTupleArray new;   /* {"columns": ...} */
 } LogicalMessageInsert;
 
 typedef struct LogicalMessageUpdate
 {
-	char nspname[PG_NAMEDATALEN];
-	char relname[PG_NAMEDATALEN];
+	LogicalMessageRelation table;
 	LogicalMessageTupleArray old;   /* {"identity": ...} */
 	LogicalMessageTupleArray new;   /* {"columns": ...} */
 } LogicalMessageUpdate;
 
 typedef struct LogicalMessageDelete
 {
-	char nspname[PG_NAMEDATALEN];
-	char relname[PG_NAMEDATALEN];
+	LogicalMessageRelation table;
 	LogicalMessageTupleArray old;   /* {"identity": ...} */
 } LogicalMessageDelete;
 
 typedef struct LogicalMessageTruncate
 {
-	char nspname[PG_NAMEDATALEN];
-	char relname[PG_NAMEDATALEN];
+	LogicalMessageRelation table;
 } LogicalMessageTruncate;
 
 typedef struct LogicalMessageSwitchWAL
@@ -206,9 +211,11 @@ typedef struct LogicalTransaction
 	uint32_t xid;
 	uint64_t beginLSN;
 	uint64_t commitLSN;
+	uint64_t rollbackLSN;
 	char timestamp[PG_MAX_TIMESTAMP];
 	bool continued;
 	bool commit;
+	bool rollback;
 
 	uint32_t count;                     /* number of statements */
 	LogicalTransactionStatement *first;
@@ -284,9 +291,6 @@ typedef struct StreamContext
 	uint64_t endpos;
 	bool apply;
 
-	bool reachedStartPos;
-	StreamAction startposActionFromJSON;
-
 	bool stdIn;
 	bool stdOut;
 
@@ -303,9 +307,11 @@ typedef struct StreamContext
 	uint64_t lastWriteTime;
 
 	/* transform needs some catalog lookups (pkey, type oid) */
-	SourceCatalog *catalog;
+	DatabaseCatalog *sourceDB;
 
 	Queue *transformQueue;
+	PGSQL *transformPGSQL;
+
 	uint32_t WalSegSz;
 	uint32_t timeline;
 
@@ -317,6 +323,8 @@ typedef struct StreamContext
 	FILE *sqlFile;
 
 	StreamCounters counters;
+
+	bool transactionInProgress;
 } StreamContext;
 
 
@@ -356,9 +364,8 @@ typedef struct StreamApplyContext
 	/* target connection */
 	PGSQL pgsql;
 
-	/* source connection to publish sentinel updates */
-	PGSQL src;
-	bool sentinelQueryInProgress;
+	/* apply needs access to the catalogs to register sentinel replay_lsn */
+	DatabaseCatalog *sourceDB;
 	uint64_t sentinelSyncTime;
 
 	ConnStrings *connStrings;
@@ -441,9 +448,6 @@ struct StreamSpecs
 	uint64_t endpos;
 	CopyDBSentinel sentinel;
 
-	bool startposComputedFromJSON;
-	StreamAction startposActionFromJSON;
-
 	LogicalStreamMode mode;
 
 	bool restart;
@@ -456,10 +460,11 @@ struct StreamSpecs
 	FollowSubProcess catchup;
 
 	/* transform needs some catalog lookups (pkey, type oid) */
-	SourceCatalog *catalog;
+	DatabaseCatalog *sourceDB;
 
 	/* receive push json filenames to a queue for transform */
 	Queue transformQueue;
+	PGSQL transformPGSQL;
 
 	/* ld_stream and ld_transform needs their own StreamContext instance */
 	StreamContext private;
@@ -484,7 +489,7 @@ bool stream_init_specs(StreamSpecs *specs,
 					   char *origin,
 					   uint64_t endpos,
 					   LogicalStreamMode mode,
-					   SourceCatalog *catalog,
+					   DatabaseCatalog *sourceDB,
 					   bool stdIn,
 					   bool stdOut,
 					   bool logSQL);
@@ -532,6 +537,8 @@ bool stream_read_latest(StreamSpecs *specs, StreamContent *content);
 bool stream_update_latest_symlink(StreamContext *privateContext,
 								  const char *filename);
 
+bool stream_sync_sentinel(LogicalStreamContext *context);
+
 bool buildReplicationURI(const char *pguri, char **repl_pguri);
 
 bool stream_setup_databases(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs);
@@ -547,6 +554,10 @@ bool stream_create_sentinel(CopyDataSpec *copySpecs,
 							uint64_t startpos,
 							uint64_t endpos);
 
+bool stream_fetch_current_lsn(uint64_t *lsn,
+							  const char *pguri,
+							  ConnectionType connectionType);
+
 bool stream_write_context(StreamSpecs *specs, LogicalStreamClient *stream);
 bool stream_cleanup_context(StreamSpecs *specs);
 bool stream_read_context(CDCPaths *paths,
@@ -569,6 +580,7 @@ bool stream_compute_pathnames(uint32_t WalSegSz,
 							  char *walFileName,
 							  char *sqlFileName);
 
+bool stream_transform_context_init_pgsql(StreamSpecs *specs);
 bool stream_transform_stream(StreamSpecs *specs);
 bool stream_transform_resume(StreamSpecs *specs);
 bool stream_transform_line(void *ctx, const char *line, bool *stop);
@@ -596,6 +608,7 @@ bool stream_write_endpos(FILE *out, LogicalMessageEndpos *endpos);
 
 bool stream_write_begin(FILE *out, LogicalTransaction *tx);
 bool stream_write_commit(FILE *out, LogicalTransaction *tx);
+bool stream_write_rollback(FILE *out, LogicalTransaction *tx);
 
 bool stream_write_insert(FILE *out, LogicalMessageInsert *insert);
 bool stream_write_truncate(FILE *out, LogicalMessageTruncate *truncate);
@@ -615,6 +628,7 @@ bool streamLogicalTransactionAppendStatement(LogicalTransaction *txn,
 void FreeLogicalMessage(LogicalMessage *msg);
 void FreeLogicalTransaction(LogicalTransaction *tx);
 void FreeLogicalMessageTupleArray(LogicalMessageTupleArray *tupleArray);
+void FreeLogicalMessageRelation(LogicalMessageRelation *table);
 void FreeLogicalMessageTuple(LogicalMessageTuple *tuple);
 bool AllocateLogicalMessageTuple(LogicalMessageTuple *tuple, int count);
 
@@ -647,9 +661,6 @@ bool stream_apply_wait_for_sentinel(StreamSpecs *specs,
 bool stream_apply_sync_sentinel(StreamApplyContext *context,
 								bool findDurableLSN);
 
-bool stream_apply_send_sync_sentinel(StreamApplyContext *context);
-bool stream_apply_fetch_sync_sentinel(StreamApplyContext *context);
-
 bool stream_apply_file(StreamApplyContext *context);
 
 bool stream_apply_sql(StreamApplyContext *context,
@@ -657,6 +668,7 @@ bool stream_apply_sql(StreamApplyContext *context,
 					  const char *sql);
 
 bool stream_apply_init_context(StreamApplyContext *context,
+							   DatabaseCatalog *sourceDB,
 							   CDCPaths *paths,
 							   ConnStrings *connStrings,
 							   char *origin,
