@@ -14,12 +14,14 @@
 
 #include "postgres.h"
 #include "postgres_fe.h"
+#include "libpq-fe.h"
 #include "access/xlog_internal.h"
 #include "access/xlogdefs.h"
 
 #include "lookup3.h"
 #include "parson.h"
 
+#include "catalog.h"
 #include "cli_common.h"
 #include "cli_root.h"
 #include "copydb.h"
@@ -30,6 +32,7 @@
 #include "parsing_utils.h"
 #include "pidfile.h"
 #include "pg_utils.h"
+#include "pgsql.h"
 #include "schema.h"
 #include "signals.h"
 #include "string_utils.h"
@@ -42,11 +45,46 @@ typedef struct TransformStreamCtx
 	uint64_t currentMsgIndex;
 } TransformStreamCtx;
 
+static bool stream_transform_stream_internal(StreamSpecs *specs);
+
+static bool stream_transform_from_queue_internal(StreamSpecs *specs);
+
 static bool canCoalesceLogicalTransactionStatement(LogicalTransaction *txn,
 												   LogicalTransactionStatement *new);
 static bool coalesceLogicalTransactionStatement(LogicalTransaction *txn,
 												LogicalTransactionStatement *new);
 
+/*
+ * stream_transform_context_init_pgsql initializes StreamContext's
+ * transformPGSQL and opens a connection to the target. This is required to use
+ * PQescapeIdentifier API of libpq when escaping identifiers
+ */
+bool
+stream_transform_context_init_pgsql(StreamSpecs *specs)
+{
+	StreamContext *privateContext = &(specs->private);
+
+	privateContext->transformPGSQL = &(specs->transformPGSQL);
+
+	/* initialize our connection to the target database */
+	if (!pgsql_init(privateContext->transformPGSQL,
+					specs->connStrings->target_pguri,
+					PGSQL_CONN_TARGET))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	if (!pgsql_open_connection(privateContext->transformPGSQL))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	return true;
+}
+
+
 /*
  * stream_transform_stream transforms a JSON formatted input stream (read line
  * by line) as received from the wal2json logical decoding plugin into an SQL
@@ -54,6 +92,27 @@ static bool coalesceLogicalTransactionStatement(LogicalTransaction *txn,
  */
 bool
 stream_transform_stream(StreamSpecs *specs)
+{
+	if (!stream_transform_context_init_pgsql(specs))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	bool success = stream_transform_stream_internal(specs);
+
+	pgsql_finish(&(specs->transformPGSQL));
+
+	return success;
+}
+
+
+/*
+ * stream_transform_stream_internal implements the core of
+ * stream_transform_stream
+ */
+static bool
+stream_transform_stream_internal(StreamSpecs *specs)
 {
 	StreamContext *privateContext = &(specs->private);
 
@@ -142,27 +201,28 @@ stream_transform_resume(StreamSpecs *specs)
 	 * command line option (found in specs->endpos) prevails, but when it's not
 	 * been used, we have a look at the sentinel value.
 	 */
-	PGSQL src = { 0 };
-
-	if (!pgsql_init(&src, specs->connStrings->source_pguri, PGSQL_CONN_SOURCE))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
 	CopyDBSentinel sentinel = { 0 };
 
-	if (!pgsql_get_sentinel(&src, &sentinel))
+	if (!sentinel_get(specs->sourceDB, &sentinel))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
+	log_notice("stream_transform_resume: "
+			   "startpos %X/%X endpos %X/%X "
+			   "write_lsn %X/%X flush_lsn %X/%X replay_lsn %X/%X",
+			   LSN_FORMAT_ARGS(sentinel.startpos),
+			   LSN_FORMAT_ARGS(sentinel.endpos),
+			   LSN_FORMAT_ARGS(sentinel.write_lsn),
+			   LSN_FORMAT_ARGS(sentinel.flush_lsn),
+			   LSN_FORMAT_ARGS(sentinel.replay_lsn));
+
 	if (specs->endpos == InvalidXLogRecPtr)
 	{
 		specs->endpos = sentinel.endpos;
 	}
-	else
+	else if (specs->endpos != sentinel.endpos)
 	{
 		log_warn("Sentinel endpos is %X/%X, overriden by --endpos %X/%X",
 				 LSN_FORMAT_ARGS(sentinel.endpos),
@@ -182,10 +242,8 @@ stream_transform_resume(StreamSpecs *specs)
 		{
 			specs->startpos = sentinel.startpos;
 
-			log_info("Resuming streaming at LSN %X/%X "
-					 "from replication slot \"%s\"",
-					 LSN_FORMAT_ARGS(specs->startpos),
-					 specs->slot.slotName);
+			log_notice("Resuming transform at LSN %X/%X from sentinel",
+					   LSN_FORMAT_ARGS(specs->startpos));
 		}
 	}
 
@@ -314,24 +372,16 @@ stream_transform_line(void *ctx, const char *line, bool *stop)
 	/* at ENDPOS check that it's the current sentinel value and exit */
 	else if (metadata->action == STREAM_ACTION_ENDPOS)
 	{
-		PGSQL src = { 0 };
-		char *dsn = privateContext->connStrings->source_pguri;
-
-		if (!pgsql_init(&src, dsn, PGSQL_CONN_SOURCE))
-		{
-			/* errors have already been logged */
-			return false;
-		}
-
 		CopyDBSentinel sentinel = { 0 };
 
-		if (!pgsql_get_sentinel(&src, &sentinel))
+		if (!sentinel_get(privateContext->sourceDB, &sentinel))
 		{
 			/* errors have already been logged */
 			return false;
 		}
 
-		if (sentinel.endpos <= metadata->lsn)
+		if (sentinel.endpos != InvalidXLogRecPtr &&
+			sentinel.endpos <= metadata->lsn)
 		{
 			*stop = true;
 
@@ -373,6 +423,7 @@ stream_transform_write_message(StreamContext *privateContext,
 	 * least a partial transaction within known boundaries.
 	 */
 	if (metadata->action != STREAM_ACTION_COMMIT &&
+		metadata->action != STREAM_ACTION_ROLLBACK &&
 		metadata->action != STREAM_ACTION_KEEPALIVE &&
 		metadata->action != STREAM_ACTION_SWITCH &&
 		metadata->action != STREAM_ACTION_ENDPOS)
@@ -407,7 +458,8 @@ stream_transform_write_message(StreamContext *privateContext,
 
 	(void) FreeLogicalMessage(currentMsg);
 
-	if (metadata->action == STREAM_ACTION_COMMIT)
+	if (metadata->action == STREAM_ACTION_COMMIT ||
+		metadata->action == STREAM_ACTION_ROLLBACK)
 	{
 		/* then prepare a new one, reusing the same memory area */
 		LogicalMessage empty = { 0 };
@@ -431,16 +483,16 @@ stream_transform_write_message(StreamContext *privateContext,
 		new.isTransaction = true;
 		new.action = STREAM_ACTION_BEGIN;
 
-		LogicalTransaction *old = &(currentMsg->command.tx);
-		LogicalTransaction *txn = &(new.command.tx);
+		LogicalTransaction *oldTxn = &(currentMsg->command.tx);
+		LogicalTransaction *newTxn = &(new.command.tx);
 
-		txn->continued = true;
+		newTxn->continued = true;
 
-		txn->xid = old->xid;
-		txn->beginLSN = old->beginLSN;
-		strlcpy(txn->timestamp, old->timestamp, sizeof(txn->timestamp));
+		newTxn->xid = oldTxn->xid;
+		newTxn->beginLSN = oldTxn->beginLSN;
+		strlcpy(newTxn->timestamp, oldTxn->timestamp, sizeof(newTxn->timestamp));
 
-		txn->first = NULL;
+		newTxn->first = NULL;
 
 		*currentMsg = new;
 	}
@@ -607,10 +659,7 @@ stream_transform_worker(StreamSpecs *specs)
 bool
 stream_transform_from_queue(StreamSpecs *specs)
 {
-	Queue *transformQueue = &(specs->transformQueue);
-
-	int errors = 0;
-	bool stop = false;
+	DatabaseCatalog *sourceDB = specs->sourceDB;
 
 	if (!stream_init_context(specs))
 	{
@@ -618,10 +667,42 @@ stream_transform_from_queue(StreamSpecs *specs)
 		return false;
 	}
 
+	if (!stream_transform_context_init_pgsql(specs))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	bool success = stream_transform_from_queue_internal(specs);
+
+	pgsql_finish(&(specs->transformPGSQL));
+
+	if (!catalog_close(sourceDB))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	return success;
+}
+
+
+/*
+ * stream_transform_from_queue_internal implements the core of
+ * stream_transform_from_queue
+ */
+static bool
+stream_transform_from_queue_internal(StreamSpecs *specs)
+{
+	Queue *transformQueue = &(specs->transformQueue);
+
+	int errors = 0;
+	bool stop = false;
+
 	while (!stop)
 	{
-		QMessage *mesg = (QMessage *) calloc(1, sizeof(QMessage));
-		bool recv_ok = queue_receive(transformQueue, mesg);
+		QMessage mesg = { 0 };
+		bool recv_ok = queue_receive(transformQueue, &mesg);
 
 		if (asked_to_stop || asked_to_stop_fast || asked_to_quit)
 		{
@@ -631,18 +712,16 @@ stream_transform_from_queue(StreamSpecs *specs)
 			 * catchup mode to replay mode.
 			 */
 			log_debug("stream_transform_from_queue was asked to stop");
-			free(mesg);
 			return true;
 		}
 
 		if (!recv_ok)
 		{
 			/* errors have already been logged */
-			free(mesg);
 			return false;
 		}
 
-		switch (mesg->type)
+		switch (mesg.type)
 		{
 			case QMSG_TYPE_STOP:
 			{
@@ -654,14 +733,13 @@ stream_transform_from_queue(StreamSpecs *specs)
 			case QMSG_TYPE_STREAM_TRANSFORM:
 			{
 				log_debug("stream_transform_from_queue: %X/%X",
-						  LSN_FORMAT_ARGS(mesg->data.lsn));
+						  LSN_FORMAT_ARGS(mesg.data.lsn));
 
-				if (!stream_transform_file_at_lsn(specs, mesg->data.lsn))
+				if (!stream_transform_file_at_lsn(specs, mesg.data.lsn))
 				{
 					/* errors have already been logged, break from the loop */
 					++errors;
-					free(mesg);
-					return false;
+					break;
 				}
 
 				break;
@@ -670,15 +748,13 @@ stream_transform_from_queue(StreamSpecs *specs)
 			default:
 			{
 				log_error("Received unknown message type %ld on %s queue %d",
-						  mesg->type,
+						  mesg.type,
 						  transformQueue->name,
 						  transformQueue->qId);
 				++errors;
 				break;
 			}
 		}
-
-		free(mesg);
 	}
 
 	bool success = (stop == true && errors == 0);
@@ -902,6 +978,7 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 		 */
 		if (firstMessage &&
 			(metadata->action == STREAM_ACTION_COMMIT ||
+			 metadata->action == STREAM_ACTION_ROLLBACK ||
 			 metadata->action == STREAM_ACTION_INSERT ||
 			 metadata->action == STREAM_ACTION_UPDATE ||
 			 metadata->action == STREAM_ACTION_DELETE ||
@@ -1065,13 +1142,15 @@ parseMessage(StreamContext *privateContext, char *message, JSON_Value *json)
 	}
 
 	/*
-	 * All messages except for BEGIN/COMMIT need a LogicalTransactionStatement
-	 * to represent them within the current transaction.
+	 * All messages except for BEGIN/COMMIT/ROLLBACL need a
+	 * LogicalTransactionStatement to represent them within the
+	 * current transaction.
 	 */
 	LogicalTransactionStatement *stmt = NULL;
 
 	if (metadata->action != STREAM_ACTION_BEGIN &&
-		metadata->action != STREAM_ACTION_COMMIT)
+		metadata->action != STREAM_ACTION_COMMIT &&
+		metadata->action != STREAM_ACTION_ROLLBACK)
 	{
 		stmt = (LogicalTransactionStatement *)
 			   calloc(1, sizeof(LogicalTransactionStatement));
@@ -1142,6 +1221,19 @@ parseMessage(StreamContext *privateContext, char *message, JSON_Value *json)
 			break;
 		}
 
+		case STREAM_ACTION_ROLLBACK:
+		{
+			if (!mesg->isTransaction)
+			{
+				log_error("Failed to parse ROLLBACK: no transaction in progress");
+				return false;
+			}
+
+			txn->rollbackLSN = metadata->lsn;
+			txn->rollback = true;
+			break;
+		}
+
 		/* switch wal messages are pgcopydb internal messages */
 		case STREAM_ACTION_SWITCH:
 		{
@@ -1365,8 +1457,8 @@ canCoalesceLogicalTransactionStatement(LogicalTransaction *txn,
 	LogicalMessageInsert *newInsert = &new->stmt.insert;
 
 	/* Last and current statements must target same relation */
-	if (!streq(lastInsert->nspname, newInsert->nspname) ||
-		!streq(lastInsert->relname, newInsert->relname))
+	if (!streq(lastInsert->table.nspname, newInsert->table.nspname) ||
+		!streq(lastInsert->table.relname, newInsert->table.relname))
 	{
 		return false;
 	}
@@ -1505,12 +1597,14 @@ FreeLogicalTransaction(LogicalTransaction *tx)
 		{
 			case STREAM_ACTION_INSERT:
 			{
+				FreeLogicalMessageRelation(&(currentStmt->stmt.insert.table));
 				FreeLogicalMessageTupleArray(&(currentStmt->stmt.insert.new));
 				break;
 			}
 
 			case STREAM_ACTION_UPDATE:
 			{
+				FreeLogicalMessageRelation(&(currentStmt->stmt.update.table));
 				FreeLogicalMessageTupleArray(&(currentStmt->stmt.update.old));
 				FreeLogicalMessageTupleArray(&(currentStmt->stmt.update.new));
 				break;
@@ -1518,10 +1612,17 @@ FreeLogicalTransaction(LogicalTransaction *tx)
 
 			case STREAM_ACTION_DELETE:
 			{
+				FreeLogicalMessageRelation(&(currentStmt->stmt.delete.table));
 				FreeLogicalMessageTupleArray(&(currentStmt->stmt.delete.old));
 				break;
 			}
 
+			case STREAM_ACTION_TRUNCATE:
+			{
+				FreeLogicalMessageRelation(&(currentStmt->stmt.truncate.table));
+				break;
+			}
+
 			/* no malloc'ated area in a BEGIN, COMMIT, or TRUNCATE statement */
 			default:
 			{
@@ -1539,6 +1640,27 @@ FreeLogicalTransaction(LogicalTransaction *tx)
 }
 
 
+/*
+ * FreeLogicalMessageRelation frees the malloc'ated memory areas of
+ * LogicalMessageRelation.
+ */
+void
+FreeLogicalMessageRelation(LogicalMessageRelation *table)
+{
+	if (table->pqMemory)
+	{
+		/* use PQfreemem for memory allocated by PQescapeIdentifer */
+		PQfreemem(table->nspname);
+		PQfreemem(table->relname);
+	}
+	else
+	{
+		free(table->nspname);
+		free(table->relname);
+	}
+}
+
+
 /*
  * FreeLogicalMessageTupleArray frees the malloc'ated memory areas of a
  * LogicalMessageTupleArray.
@@ -1888,6 +2010,14 @@ stream_write_transaction(FILE *out, LogicalTransaction *txn)
 		}
 	}
 
+	if (txn->rollback)
+	{
+		if (!stream_write_rollback(out, txn))
+		{
+			return false;
+		}
+	}
+
 	/* flush out stream at transaction boundaries */
 	if (fflush(out) != 0)
 	{
@@ -1937,6 +2067,24 @@ stream_write_begin(FILE *out, LogicalTransaction *txn)
 }
 
 
+/*
+ * stream_write_rollback writes a COMMIT statement to the already open out
+ * stream.
+ */
+bool
+stream_write_rollback(FILE *out, LogicalTransaction *txn)
+{
+	FFORMAT(out,
+			"%s{\"xid\":%lld,\"lsn\":\"%X/%X\",\"timestamp\":\"%s\"}\n",
+			OUTPUT_ROLLBACK,
+			(long long) txn->xid,
+			LSN_FORMAT_ARGS(txn->rollbackLSN),
+			txn->timestamp);
+
+	return true;
+}
+
+
 /*
  * stream_write_commit writes a COMMIT statement to the already open out
  * stream.
@@ -2022,8 +2170,8 @@ stream_write_insert(FILE *out, LogicalMessageInsert *insert)
 		 * First, the PREPARE part.
 		 */
 		appendPQExpBuffer(buf, "INSERT INTO %s.%s ",
-						  insert->nspname,
-						  insert->relname);
+						  insert->table.nspname,
+						  insert->table.relname);
 
 		/* loop over column names and add them to the out stream */
 		appendPQExpBuffer(buf, "%s", "(");
@@ -2159,8 +2307,8 @@ stream_write_update(FILE *out, LogicalMessageUpdate *update)
 		 * First, the PREPARE part.
 		 */
 		appendPQExpBuffer(buf, "UPDATE %s.%s SET ",
-						  update->nspname,
-						  update->relname);
+						  update->table.nspname,
+						  update->table.relname);
 		int pos = 0;
 
 		for (int r = 0; r < new->values.count; r++)
@@ -2327,8 +2475,8 @@ stream_write_delete(FILE *out, LogicalMessageDelete *delete)
 		 * First, the PREPARE part.
 		 */
 		appendPQExpBuffer(buf, "DELETE FROM %s.%s WHERE ",
-						  delete->nspname,
-						  delete->relname);
+						  delete->table.nspname,
+						  delete->table.relname);
 
 		int pos = 0;
 
@@ -2400,7 +2548,10 @@ stream_write_delete(FILE *out, LogicalMessageDelete *delete)
 bool
 stream_write_truncate(FILE *out, LogicalMessageTruncate *truncate)
 {
-	FFORMAT(out, "TRUNCATE ONLY %s.%s\n", truncate->nspname, truncate->relname);
+	FFORMAT(out,
+			"TRUNCATE ONLY %s.%s\n",
+			truncate->table.nspname,
+			truncate->table.relname);
 
 	return true;
 }
