@@ -11,6 +11,7 @@
 
 #include "postgres.h"
 #include "postgres_fe.h"
+#include "libpq-fe.h"
 #include "access/xlog_internal.h"
 #include "access/xlogdefs.h"
 
@@ -24,6 +25,7 @@
 #include "lock_utils.h"
 #include "log.h"
 #include "parsing_utils.h"
+#include "pgsql.h"
 #include "pidfile.h"
 #include "pg_utils.h"
 #include "schema.h"
@@ -32,6 +34,9 @@
 #include "summary.h"
 
 
+static bool SetMessageRelation(JSON_Object *jsobj,
+							   LogicalMessageRelation *table,
+							   PGSQL *pgsql);
 static bool SetColumnNamesAndValues(LogicalMessageTuple *tuple,
 									const char *message,
 									JSON_Array *jscols);
@@ -120,18 +125,13 @@ parseWal2jsonMessage(StreamContext *privateContext,
 	/* most actions share a need for "schema" and "table" properties */
 	JSON_Object *jsobj = json_value_get_object(json);
 
-	char *schema = NULL;
-	char *table = NULL;
+	LogicalMessageRelation table = { 0 };
 
-	schema = (char *) json_object_dotget_string(jsobj, "message.schema");
-	table = (char *) json_object_dotget_string(jsobj, "message.table");
-
-	if (schema == NULL || table == NULL)
+	if (!SetMessageRelation(jsobj, &table, privateContext->transformPGSQL))
 	{
 		log_error("Failed to parse truncated message missing "
 				  "schema or table property: %s",
 				  message);
-		return false;
 	}
 
 	switch (metadata->action)
@@ -148,9 +148,7 @@ parseWal2jsonMessage(StreamContext *privateContext,
 
 		case STREAM_ACTION_TRUNCATE:
 		{
-			strlcpy(stmt->stmt.truncate.nspname, schema, PG_NAMEDATALEN);
-			strlcpy(stmt->stmt.truncate.relname, table, PG_NAMEDATALEN);
-
+			stmt->stmt.truncate.table = table;
 			break;
 		}
 
@@ -159,8 +157,7 @@ parseWal2jsonMessage(StreamContext *privateContext,
 			JSON_Array *jscols =
 				json_object_dotget_array(jsobj, "message.columns");
 
-			strlcpy(stmt->stmt.insert.nspname, schema, PG_NAMEDATALEN);
-			strlcpy(stmt->stmt.insert.relname, table, PG_NAMEDATALEN);
+			stmt->stmt.insert.table = table;
 
 			stmt->stmt.insert.new.count = 1;
 			stmt->stmt.insert.new.array =
@@ -187,8 +184,7 @@ parseWal2jsonMessage(StreamContext *privateContext,
 
 		case STREAM_ACTION_UPDATE:
 		{
-			strlcpy(stmt->stmt.update.nspname, schema, PG_NAMEDATALEN);
-			strlcpy(stmt->stmt.update.relname, table, PG_NAMEDATALEN);
+			stmt->stmt.update.table = table;
 
 			stmt->stmt.update.old.count = 1;
 			stmt->stmt.update.new.count = 1;
@@ -235,8 +231,7 @@ parseWal2jsonMessage(StreamContext *privateContext,
 
 		case STREAM_ACTION_DELETE:
 		{
-			strlcpy(stmt->stmt.delete.nspname, schema, PG_NAMEDATALEN);
-			strlcpy(stmt->stmt.delete.relname, table, PG_NAMEDATALEN);
+			stmt->stmt.delete.table = table;
 
 			stmt->stmt.delete.old.count = 1;
 			stmt->stmt.delete.old.array =
@@ -275,6 +270,46 @@ parseWal2jsonMessage(StreamContext *privateContext,
 }
 
 
+/*
+ * SetMessageRelation parses the table's nspname and relname from the JSON
+ * object and escapes it appropriately to be put as it is in SQL statements
+ */
+static bool
+SetMessageRelation(JSON_Object *jsobj,
+				   LogicalMessageRelation *table,
+				   PGSQL *pgsql)
+{
+	char *schema = NULL;
+	char *relname = NULL;
+
+	schema = (char *) json_object_dotget_string(jsobj, "message.schema");
+	relname = (char *) json_object_dotget_string(jsobj, "message.table");
+
+
+	if (schema == NULL || relname == NULL)
+	{
+		return false;
+	}
+
+	table->nspname = pgsql_escape_identifier(pgsql, schema);
+	if (table->nspname == NULL)
+	{
+		return false;
+	}
+
+	table->relname = pgsql_escape_identifier(pgsql, relname);
+	if (table->relname == NULL)
+	{
+		PQfreemem(table->nspname);
+		return false;
+	}
+
+	table->pqMemory = true;
+
+	return true;
+}
+
+
 /*
  * SetColumnNames parses the "columns" (or "identity") JSON object from a
  * wal2json logical replication message and fills-in our internal
