@@ -75,15 +75,6 @@ main(int argc, char **argv)
 	/* register our System V resources clean-up atexit */
 	atexit(unlink_system_res_atexit);
 
-	/*
-	 * When PGCOPYDB_DEBUG is set in the environment, provide the user
-	 * commands available to debug a pgcopydb instance.
-	 */
-	if (env_exists(PGCOPYDB_DEBUG))
-	{
-		command = root_with_debug;
-	}
-
 	/*
 	 * We need to follow POSIX specifications for argument parsing, in
 	 * particular we want getopt() to stop as soon as it reaches a non option
@@ -264,6 +255,13 @@ set_logger()
 static void
 unlink_system_res_atexit(void)
 {
+	(void) copydb_cleanup_sysv_resources(&system_res_array);
+
+	/*
+	 * close the log file after cleaning up System V resources. This ordering
+	 * is important because we want to log any errors that might happen during
+	 * System V resources clean-up.
+	 */
 	if (logfp != NULL)
 	{
 		if (fclose(logfp) != 0)
@@ -271,6 +269,4 @@ unlink_system_res_atexit(void)
 			fformat(stderr, "Failed to close log file: %m\n");
 		}
 	}
-
-	(void) copydb_cleanup_sysv_resources(&system_res_array);
 }
