@@ -15,6 +15,7 @@
 #include "postgres_fe.h"
 #include "pqexpbuffer.h"
 
+#include "catalog.h"
 #include "cli_root.h"
 #include "defaults.h"
 #include "env_utils.h"
@@ -30,6 +31,9 @@
 #define RUN_PROGRAM_IMPLEMENTATION
 #include "runprogram.h"
 
+static bool pg_dump_db_extension_namespace_hook(void *ctx, SourceExtension *ext);
+
+
 /*
  * Get psql --version output in pgPaths->pg_version.
  */
@@ -335,6 +339,13 @@ set_psql_from_pg_config(PostgresPaths *pgPaths)
 }
 
 
+typedef struct DumpExtensionNamespaceContext
+{
+	char **extNamespaces;
+	int *extNamespaceCount;
+} DumpExtensionNamespaceContext;
+
+
 /*
  * Call pg_dump and get the given section of the dump into the target file.
  */
@@ -344,7 +355,7 @@ pg_dump_db(PostgresPaths *pgPaths,
 		   const char *snapshot,
 		   const char *section,
 		   SourceFilters *filters,
-		   SourceExtensionArray *extensionArray,
+		   DatabaseCatalog *filtersDB,
 		   const char *filename)
 {
 	char *args[PG_CMD_MAX_ARG];
@@ -420,30 +431,44 @@ pg_dump_db(PostgresPaths *pgPaths,
 		args[argsIndex++] = nspname;
 	}
 
+	/*
+	 * Store extension args in a separate array, extension args will dynamically
+	 * allocated by pg_dump_db_extension_namespace_hook and we want to free
+	 * them after the call to pg_dump.
+	 */
+	char *extNamespaces[PG_CMD_MAX_ARG];
+	int extNamespaceCount = 0;
+
 	/* now --exclude-schema for extension's own schemas */
-	if (extensionArray != NULL)
+	DumpExtensionNamespaceContext context = {
+		.extNamespaces = extNamespaces,
+		.extNamespaceCount = &extNamespaceCount,
+	};
+
+	if (!catalog_iter_s_extension(filtersDB,
+								  &context,
+								  &pg_dump_db_extension_namespace_hook))
 	{
-		for (int i = 0; i < extensionArray->count; i++)
+		log_error("Failed to prepare pg_dump command line arguments, "
+				  "see above for details");
+		return false;
+	}
+
+	for (int i = 0; i < extNamespaceCount; i++)
+	{
+		/* check that we still have room for --exclude-schema args */
+		if (PG_CMD_MAX_ARG < (argsIndex + 2))
 		{
-			char *nspname = extensionArray->array[i].extnamespace;
-
-			if (!streq(nspname, "public") && !streq(nspname, "pg_catalog"))
-			{
-				/* check that we still have room for --exclude-schema args */
-				if (PG_CMD_MAX_ARG < (argsIndex + 2))
-				{
-					log_error("Failed to call pg_dump, "
-							  "too many schema are excluded: "
-							  "argsIndex %d > %d",
-							  argsIndex + 2,
-							  PG_CMD_MAX_ARG);
-					return false;
-				}
-
-				args[argsIndex++] = "--exclude-schema";
-				args[argsIndex++] = nspname;
-			}
+			log_error("Failed to call pg_dump, "
+					  "too many schema are excluded: "
+					  "argsIndex %d > %d",
+					  argsIndex + 2,
+					  PG_CMD_MAX_ARG);
+			return false;
 		}
+
+		args[argsIndex++] = "--exclude-schema";
+		args[argsIndex++] = extNamespaces[i];
 	}
 
 	args[argsIndex++] = "--file";
@@ -460,6 +485,12 @@ pg_dump_db(PostgresPaths *pgPaths,
 	(void) initialize_program(&program, args, false);
 	program.processBuffer = &processBufferCallback;
 
+	/* free the extension namespace string values */
+	for (int i = 0; i < extNamespaceCount; i++)
+	{
+		free(extNamespaces[i]);
+	}
+
 	/* log the exact command line we're using */
 	int commandSize = snprintf_program_command_line(&program, command, BUFSIZE);
 
@@ -495,6 +526,39 @@ pg_dump_db(PostgresPaths *pgPaths,
 }
 
 
+/*
+ * pg_dump_db_extension_namespace_hook is an iterator callback function.
+ */
+static bool
+pg_dump_db_extension_namespace_hook(void *ctx, SourceExtension *ext)
+{
+	DumpExtensionNamespaceContext *context =
+		(DumpExtensionNamespaceContext *) ctx;
+
+	char **extNamespaces = context->extNamespaces;
+
+	char *nspname = ext->extnamespace;
+
+	if (!streq(nspname, "public") && !streq(nspname, "pg_catalog"))
+	{
+		/* check that we still have room for args */
+		if (PG_CMD_MAX_ARG < ((*context->extNamespaceCount) + 1))
+		{
+			log_error("Failed to call pg_dump, "
+					  "too many schema are excluded: "
+					  "extNamespaceCount %d > %d",
+					  (*context->extNamespaceCount) + 1,
+					  PG_CMD_MAX_ARG);
+			return false;
+		}
+
+		extNamespaces[(*context->extNamespaceCount)++] = strdup(nspname);
+	}
+
+	return true;
+}
+
+
 /*
  * Call pg_dump and get the given section of the dump into the target file.
  */
@@ -819,7 +883,16 @@ pg_restore_db(PostgresPaths *pgPaths,
 	args[argsIndex++] = (char *) pgPaths->pg_restore;
 	args[argsIndex++] = "--dbname";
 	args[argsIndex++] = (char *) connStrings->safeTargetPGURI.pguri;
-	args[argsIndex++] = "--single-transaction";
+
+	if (options.jobs == 1)
+	{
+		args[argsIndex++] = "--single-transaction";
+	}
+	else
+	{
+		args[argsIndex++] = "--jobs";
+		args[argsIndex++] = intToString(options.jobs).strValue;
+	}
 
 	if (options.dropIfExists)
 	{
