@@ -186,6 +186,7 @@ typedef struct PGSQL
 	bool notificationReceived;
 
 	bool logSQL;
+	bool singleRowMode;
 } PGSQL;
 
 
@@ -237,6 +238,8 @@ typedef struct GUC
 
 bool pgsql_init(PGSQL *pgsql, char *url, ConnectionType connectionType);
 
+PGconn * pgsql_open_connection(PGSQL *pgsql);
+
 void pgsql_set_retry_policy(ConnectionRetryPolicy *retryPolicy,
 							int maxT,
 							int maxR,
@@ -307,14 +310,25 @@ bool validate_connection_string(const char *connectionString);
 
 bool pgsql_truncate(PGSQL *pgsql, const char *qname);
 
-bool pg_copy(PGSQL *src, PGSQL *dst,
-			 const char *srcQname, const char *dstQname, bool truncate);
+typedef struct CopyArgs
+{
+	char *srcQname;
+	char *srcAttrList;
+	char *srcWhereClause;
+	char *dstQname;
+	char *dstAttrList;
+	bool truncate;
+	bool freeze;
+	uint64_t bytesTransmitted;
+} CopyArgs;
+
+bool pg_copy(PGSQL *src, PGSQL *dst, CopyArgs *args);
 
 bool pg_copy_from_stdin(PGSQL *pgsql, const char *qname);
 bool pg_copy_row_from_stdin(PGSQL *pgsql, char *fmt, ...);
 bool pg_copy_end(PGSQL *pgsql);
 
-bool pgsql_get_sequence(PGSQL *pgsql, const char *nspname, const char *relname,
+bool pgsql_get_sequence(PGSQL *pgsql, const char *qname,
 						int64_t *lastValue,
 						bool *isCalled);
 
@@ -323,7 +337,8 @@ bool pgsql_set_gucs(PGSQL *pgsql, GUC *settings);
 bool pg_copy_large_object(PGSQL *src,
 						  PGSQL *dst,
 						  bool dropIfExists,
-						  uint32_t oid);
+						  uint32_t oid,
+						  uint64_t *bytesTransmitted);
 
 /*
  * Maximum length of serialized pg_lsn value
@@ -475,6 +490,7 @@ typedef struct ReplicationSlot
 	uint64_t lsn;
 	char snapshot[BUFSIZE];
 	StreamOutputPlugin plugin;
+	bool wal2jsonNumericAsString;
 } ReplicationSlot;
 
 bool pgsql_create_logical_replication_slot(LogicalStreamClient *client,
@@ -529,40 +545,5 @@ bool pgsql_table_exists(PGSQL *pgsql,
 bool pgsql_current_wal_flush_lsn(PGSQL *pgsql, uint64_t *lsn);
 bool pgsql_current_wal_insert_lsn(PGSQL *pgsql, uint64_t *lsn);
 
-/*
- * pgcopydb sentinel is a table that's created on the source database and
- * allows communicating elements from the outside, and in between the receive
- * and apply processes.
- */
-typedef struct CopyDBSentinel
-{
-	bool apply;
-	uint64_t startpos;
-	uint64_t endpos;
-	uint64_t write_lsn;
-	uint64_t flush_lsn;
-	uint64_t replay_lsn;
-} CopyDBSentinel;
-
-bool pgsql_update_sentinel_startpos(PGSQL *pgsql, uint64_t startpos);
-bool pgsql_update_sentinel_endpos(PGSQL *pgsql, bool current, uint64_t endpos);
-bool pgsql_update_sentinel_apply(PGSQL *pgsql, bool apply);
-
-bool pgsql_get_sentinel(PGSQL *pgsql, CopyDBSentinel *sentinel);
-
-bool pgsql_sync_sentinel_recv(PGSQL *pgsql,
-							  uint64_t write_lsn,
-							  uint64_t flush_lsn,
-							  CopyDBSentinel *sentinel);
-
-bool pgsql_sync_sentinel_apply(PGSQL *pgsql,
-							   uint64_t replay_lsn,
-							   CopyDBSentinel *sentinel);
-
-bool pgsql_send_sync_sentinel_apply(PGSQL *pgsql, uint64_t replay_lsn);
-bool pgsql_fetch_sync_sentinel_apply(PGSQL *pgsql,
-									 bool *retry,
-									 CopyDBSentinel *sentinel);
-
-
+char * pgsql_escape_identifier(PGSQL *pgsql, char *src);
 #endif /* PGSQL_H */
