@@ -9,6 +9,7 @@
 #include <sys/wait.h>
 #include <unistd.h>
 
+#include "catalog.h"
 #include "copydb.h"
 #include "env_utils.h"
 #include "lock_utils.h"
@@ -20,6 +21,16 @@
 #include "summary.h"
 
 
+static bool copydb_prepare_sequence_specs_hook(void *ctx, SourceSequence *seq);
+static bool copydb_copy_all_sequences_hook(void *ctx, SourceSequence *seq);
+
+typedef struct PrepareSequenceContext
+{
+	PGSQL *pgsql;
+	DatabaseCatalog *sourceDB;
+} PrepareSequenceContext;
+
+
 /*
  * sequence_prepare_specs fetches the list of sequences at pgsql connection,
  * using the filtering already prepared in the connection (as temp tables).
@@ -28,69 +39,109 @@
 bool
 copydb_prepare_sequence_specs(CopyDataSpec *specs, PGSQL *pgsql)
 {
-	SourceSequenceArray *sequenceArray = &(specs->catalog.sequenceArray);
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
 
-	if (!schema_list_sequences(pgsql, &(specs->filters), sequenceArray))
+	TopLevelTiming timing = {
+		.label = CopyDataSectionToString(DATA_SECTION_SET_SEQUENCES)
+	};
+
+	(void) catalog_start_timing(&timing);
+
+	if (!schema_list_sequences(pgsql, &(specs->filters), sourceDB))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	SourceSequence *sourceSeqHashByOid = NULL;
+	CatalogCounts count = { 0 };
 
-	log_info("Fetching information for %d sequences", sequenceArray->count);
-
-	int errors = 0;
-
-	for (int seqIndex = 0; seqIndex < sequenceArray->count; seqIndex++)
+	if (!catalog_count_objects(sourceDB, &count))
 	{
-		SourceSequence *seq = &(sequenceArray->array[seqIndex]);
-
-		/* add the current sequence to the sequence Hash-by-OID */
-		HASH_ADD(hh, sourceSeqHashByOid, oid, sizeof(uint32_t), seq);
-
-		/*
-		 * In case of "permission denied" for SELECT on the sequence object, we
-		 * would then have a broken transaction and all the rest of the loop
-		 * would get the following:
-		 *
-		 * ERROR: current transaction is aborted, commands ignored
-		 * until end of transaction block
-		 *
-		 * To avoid that, for each sequence we first see if we're granted the
-		 * SELECT privilege.
-		 */
-		bool granted = false;
-
-		if (!pgsql_has_sequence_privilege(pgsql, seq->qname, "select", &granted))
-		{
-			/* errors have been logged */
-			++errors;
-			break;
-		}
-
-		if (!granted)
-		{
-			log_error("Failed to SELECT values for sequence %s: "
-					  "permission denied",
-					  seq->qname);
-			++errors;
-			continue;
-		}
-
-		if (!schema_get_sequence_value(pgsql, seq))
-		{
-			/* just skip this one */
-			log_warn("Failed to get sequence values for %s", seq->qname);
-			++errors;
-			continue;
-		}
+		log_error("Failed to count objects in our catalogs");
+		return false;
 	}
 
-	/* now attach the final hash table head to the specs */
-	specs->catalog.sourceSeqHashByOid = sourceSeqHashByOid;
+	log_info("Fetching information for %lld sequences",
+			 (long long) count.sequences);
 
-	return errors == 0;
+	PrepareSequenceContext context = {
+		.pgsql = pgsql,
+		.sourceDB = sourceDB
+	};
+
+	if (!catalog_iter_s_seq(sourceDB,
+							&context,
+							&copydb_prepare_sequence_specs_hook))
+	{
+		log_error("Failed to prepare our internal sequence catalogs, "
+				  "see above for details");
+		return false;
+	}
+
+	(void) catalog_stop_timing(&timing);
+
+	if (!catalog_register_section(sourceDB, &timing))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	return true;
+}
+
+
+/*
+ * copydb_prepare_sequence_specs_hook is an iterator callback function.
+ */
+static bool
+copydb_prepare_sequence_specs_hook(void *ctx, SourceSequence *seq)
+{
+	PrepareSequenceContext *context = (PrepareSequenceContext *) ctx;
+	PGSQL *pgsql = context->pgsql;
+
+	/*
+	 * In case of "permission denied" for SELECT on the sequence object, we
+	 * would then have a broken transaction and all the rest of the loop
+	 * would get the following:
+	 *
+	 * ERROR: current transaction is aborted, commands ignored
+	 * until end of transaction block
+	 *
+	 * To avoid that, for each sequence we first see if we're granted the
+	 * SELECT privilege.
+	 */
+	bool granted = false;
+
+	if (!pgsql_has_sequence_privilege(pgsql, seq->qname, "select", &granted))
+	{
+		/* errors have been logged */
+		return false;
+	}
+
+	if (!granted)
+	{
+		log_error("Failed to SELECT values for sequence %s: "
+				  "permission denied",
+				  seq->qname);
+		return false;
+	}
+
+	if (!schema_get_sequence_value(pgsql, seq))
+	{
+		/* just skip this one */
+		log_warn("Failed to get sequence values for %s", seq->qname);
+		return false;
+	}
+
+	if (!catalog_update_sequence_values(context->sourceDB, seq))
+	{
+		log_error("Failed to update sequences values for %s "
+				  "in our internal catalogs",
+				  seq->qname);
+		return false;
+	}
+
+	return true;
 }
 
 
@@ -145,6 +196,13 @@ copydb_start_seq_process(CopyDataSpec *specs)
 }
 
 
+typedef struct CopySeqContext
+{
+	PGSQL *dst;
+	uint64_t count;
+} CopySeqContext;
+
+
 /*
  * copydb_copy_all_sequences fetches the list of sequences from the source
  * database and then for each of them runs a SELECT last_value, is_called FROM
@@ -156,12 +214,27 @@ copydb_copy_all_sequences(CopyDataSpec *specs)
 {
 	log_notice("Now starting setval process %d [%d]", getpid(), getppid());
 
-	if (specs->dirState.sequenceCopyIsDone)
+	if (specs->runState.sequenceCopyIsDone)
 	{
 		log_info("Skipping sequences, already done on a previous run");
 		return true;
 	}
 
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
+
+	if (!catalog_open(sourceDB))
+	{
+		log_error("Failed to open internal catalogs in sequence reset worker, "
+				  "see above for details");
+		return false;
+	}
+
+	if (!summary_start_timing(sourceDB, TIMING_SECTION_SET_SEQUENCES))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	if (specs->section != DATA_SECTION_SET_SEQUENCES &&
 		specs->section != DATA_SECTION_ALL)
 	{
@@ -185,73 +258,61 @@ copydb_copy_all_sequences(CopyDataSpec *specs)
 		return false;
 	}
 
-	int errors = 0;
+	CopySeqContext context = { .dst = &dst, .count = 0 };
 
-	SourceSequenceArray *sequenceArray = &(specs->catalog.sequenceArray);
-
-	for (int seqIndex = 0; seqIndex < sequenceArray->count; seqIndex++)
+	if (!catalog_iter_s_seq(sourceDB, &context, &copydb_copy_all_sequences_hook))
 	{
-		SourceSequence *seq = &(sequenceArray->array[seqIndex]);
-
-		char qname[BUFSIZE] = { 0 };
-
-		sformat(qname, sizeof(qname), "%s.%s",
-				seq->nspname,
-				seq->relname);
-
-		if (!pgsql_savepoint(&dst, "sequences"))
-		{
-			/* errors have already been logged */
-			return false;
-		}
-
-		if (!schema_set_sequence_value(&dst, seq))
-		{
-			log_error("Failed to set sequence values for %s", qname);
-
-			if (specs->failFast)
-			{
-				(void) pgsql_commit(&dst);
-				return false;
-			}
-
-			if (!pgsql_rollback_to_savepoint(&dst, "sequences"))
-			{
-				/* errors have already been logged */
-				return false;
-			}
-
-			++errors;
-			continue;
-		}
-
-		if (!pgsql_release_savepoint(&dst, "sequences"))
-		{
-			/* errors have already been logged */
-			return false;
-		}
+		log_error("Failed to copy sequences values from our internal catalogs, "
+				  "see above for details");
+		(void) pgsql_finish(&dst);
+		return false;
 	}
 
 	if (!pgsql_commit(&dst))
 	{
 		/* errors have already been logged */
-		++errors;
+		return false;
 	}
 
-	/* and write that we successfully finished copying all sequences */
-	if (!write_file("", 0, specs->cfPaths.done.sequences))
+	if (!summary_stop_timing(sourceDB, TIMING_SECTION_SET_SEQUENCES))
 	{
-		log_warn("Failed to write the tracking file \%s\"",
-				 specs->cfPaths.done.sequences);
+		/* errors have already been logged */
+		return false;
 	}
 
-	if (errors > 0)
+	if (!summary_set_timing_count(sourceDB,
+								  TIMING_SECTION_SET_SEQUENCES,
+								  context.count))
 	{
-		log_warn("Failed to set values for %d sequences, "
-				 "see above for details",
-				 errors);
+		/* errors have already been logged */
+		return false;
+	}
+
+	if (!catalog_close(sourceDB))
+	{
+		/* errors have already been logged */
 		return false;
 	}
 
 	return true;
 }
+
+
+/*
+ * copydb_copy_all_sequences_hook is an iterator callback function.
+ */
+static bool
+copydb_copy_all_sequences_hook(void *ctx, SourceSequence *seq)
+{
+	CopySeqContext *context = (CopySeqContext *) ctx;
+
+	if (!schema_set_sequence_value(context->dst, seq))
+	{
+		log_error("Failed to set sequence values for %s", seq->qname);
+		return false;
+	}
+
+	++(context->count);
+
+	return true;
+}
