@@ -1,11 +1,11 @@
--- KEEPALIVE {"lsn":"0/2448D70","timestamp":"2023-09-13 05:26:07.779433+0000"}
-BEGIN; -- {"xid":491,"lsn":"0/244BA20","timestamp":"2023-09-13 05:26:07.780243+0000","commit_lsn":"0/244DDA0"}
+-- KEEPALIVE {"lsn":"0/24489B0","timestamp":"2023-12-21 17:23:21.330234+0000"}
+BEGIN; -- {"xid":491,"lsn":"0/244B698","timestamp":"2023-12-21 17:23:21.348819+0000","commit_lsn":"0/244DA18"}
 PREPARE d33a643f AS INSERT INTO public.rental ("rental_id", "rental_date", "inventory_id", "customer_id", "return_date", "staff_id", "last_update") overriding system value VALUES ($1, $2, $3, $4, $5, $6, $7), ($8, $9, $10, $11, $12, $13, $14);
 EXECUTE d33a643f["16050","2022-06-01 00:00:00+00","371","291",null,"1","2022-06-01 00:00:00+00","16051","2022-06-01 00:00:00+00","373","293",null,"2","2022-06-01 00:00:00+00"];
 PREPARE b05a8353 AS INSERT INTO public.payment_p2022_06 ("payment_id", "customer_id", "staff_id", "rental_id", "amount", "payment_date") overriding system value VALUES ($1, $2, $3, $4, $5, $6), ($7, $8, $9, $10, $11, $12);
 EXECUTE b05a8353["32099","291","1","16050","5.99","2022-06-01 00:00:00+00","32100","293","2","16051","5.99","2022-06-01 00:00:00+00"];
-COMMIT; -- {"xid":491,"lsn":"0/244DDA0","timestamp":"2023-09-13 05:26:07.780243+0000"}
-BEGIN; -- {"xid":492,"lsn":"0/244DDA0","timestamp":"2023-09-13 05:26:07.780588+0000","commit_lsn":"0/244EE88"}
+COMMIT; -- {"xid":491,"lsn":"0/244DA18","timestamp":"2023-12-21 17:23:21.348819+0000"}
+BEGIN; -- {"xid":492,"lsn":"0/244DA18","timestamp":"2023-12-21 17:23:21.349405+0000","commit_lsn":"0/244EB00"}
 PREPARE 32de52b9 AS UPDATE public.payment_p2022_02 SET "amount" = $1 WHERE "payment_id" = $2 and "customer_id" = $3 and "staff_id" = $4 and "rental_id" = $5 and "amount" = $6 and "payment_date" = $7;
 EXECUTE 32de52b9["11.95","23757","116","2","14763","11.99","2022-02-11 03:52:25.634006+00"];
 PREPARE 32de52b9 AS UPDATE public.payment_p2022_02 SET "amount" = $1 WHERE "payment_id" = $2 and "customer_id" = $3 and "staff_id" = $4 and "rental_id" = $5 and "amount" = $6 and "payment_date" = $7;
@@ -26,8 +26,8 @@ PREPARE 3b977bd8 AS UPDATE public.payment_p2022_07 SET "amount" = $1 WHERE "paym
 EXECUTE 3b977bd8["11.95","28814","592","1","3973","11.99","2022-07-06 12:15:38.928947+00"];
 PREPARE 3b977bd8 AS UPDATE public.payment_p2022_07 SET "amount" = $1 WHERE "payment_id" = $2 and "customer_id" = $3 and "staff_id" = $4 and "rental_id" = $5 and "amount" = $6 and "payment_date" = $7;
 EXECUTE 3b977bd8["11.95","29136","13","2","8831","11.99","2022-07-22 16:15:40.797771+00"];
-COMMIT; -- {"xid":492,"lsn":"0/244EE88","timestamp":"2023-09-13 05:26:07.780588+0000"}
-BEGIN; -- {"xid":493,"lsn":"0/244F048","timestamp":"2023-09-13 05:26:07.780640+0000","commit_lsn":"0/244F1F8"}
+COMMIT; -- {"xid":492,"lsn":"0/244EB00","timestamp":"2023-12-21 17:23:21.349405+0000"}
+BEGIN; -- {"xid":493,"lsn":"0/244ECC0","timestamp":"2023-12-21 17:23:21.349496+0000","commit_lsn":"0/244EE70"}
 PREPARE 2fa3c9c9 AS DELETE FROM public.payment_p2022_06 WHERE "payment_id" = $1 and "customer_id" = $2 and "staff_id" = $3 and "rental_id" = $4 and "amount" = $5 and "payment_date" = $6;
 EXECUTE 2fa3c9c9["32099","291","1","16050","5.99","2022-06-01 00:00:00+00"];
 PREPARE 2fa3c9c9 AS DELETE FROM public.payment_p2022_06 WHERE "payment_id" = $1 and "customer_id" = $2 and "staff_id" = $3 and "rental_id" = $4 and "amount" = $5 and "payment_date" = $6;
@@ -36,8 +36,8 @@ PREPARE 4f0082a0 AS DELETE FROM public.rental WHERE "rental_id" = $1;
 EXECUTE 4f0082a0["16050"];
 PREPARE 4f0082a0 AS DELETE FROM public.rental WHERE "rental_id" = $1;
 EXECUTE 4f0082a0["16051"];
-COMMIT; -- {"xid":493,"lsn":"0/244F1F8","timestamp":"2023-09-13 05:26:07.780640+0000"}
-BEGIN; -- {"xid":494,"lsn":"0/244F1F8","timestamp":"2023-09-13 05:26:07.780733+0000","commit_lsn":"0/244F778"}
+COMMIT; -- {"xid":493,"lsn":"0/244EE70","timestamp":"2023-12-21 17:23:21.349496+0000"}
+BEGIN; -- {"xid":494,"lsn":"0/244EE70","timestamp":"2023-12-21 17:23:21.349657+0000","commit_lsn":"0/244F3F0"}
 PREPARE 32de52b9 AS UPDATE public.payment_p2022_02 SET "amount" = $1 WHERE "payment_id" = $2 and "customer_id" = $3 and "staff_id" = $4 and "rental_id" = $5 and "amount" = $6 and "payment_date" = $7;
 EXECUTE 32de52b9["11.99","23757","116","2","14763","11.95","2022-02-11 03:52:25.634006+00"];
 PREPARE 32de52b9 AS UPDATE public.payment_p2022_02 SET "amount" = $1 WHERE "payment_id" = $2 and "customer_id" = $3 and "staff_id" = $4 and "rental_id" = $5 and "amount" = $6 and "payment_date" = $7;
@@ -58,6 +58,6 @@ PREPARE 3b977bd8 AS UPDATE public.payment_p2022_07 SET "amount" = $1 WHERE "paym
 EXECUTE 3b977bd8["11.99","28814","592","1","3973","11.95","2022-07-06 12:15:38.928947+00"];
 PREPARE 3b977bd8 AS UPDATE public.payment_p2022_07 SET "amount" = $1 WHERE "payment_id" = $2 and "customer_id" = $3 and "staff_id" = $4 and "rental_id" = $5 and "amount" = $6 and "payment_date" = $7;
 EXECUTE 3b977bd8["11.99","29136","13","2","8831","11.95","2022-07-22 16:15:40.797771+00"];
-COMMIT; -- {"xid":494,"lsn":"0/244F778","timestamp":"2023-09-13 05:26:07.780733+0000"}
--- KEEPALIVE {"lsn":"0/244F778","timestamp":"2023-09-13 05:26:07.780777+0000"}
--- ENDPOS {"lsn":"0/244F778"}
+COMMIT; -- {"xid":494,"lsn":"0/244F3F0","timestamp":"2023-12-21 17:23:21.349657+0000"}
+-- KEEPALIVE {"lsn":"0/244F3F0","timestamp":"2023-12-21 17:23:21.349789+0000"}
+-- ENDPOS {"lsn":"0/244F3F0"}
