@@ -10,11 +10,7 @@ set -e
 #  - PGCOPYDB_TABLE_JOBS
 #  - PGCOPYDB_INDEX_JOBS
 
-#
-# pgcopydb list extensions include a retry loop, so we use that as a proxy
-# to depend on the source/target Postgres images to be ready
-#
-pgcopydb list extensions --source ${PGCOPYDB_SOURCE_PGURI}
+pgcopydb ping
 
 #
 # Only start injecting DML traffic on the source database when the pagila
@@ -22,8 +18,7 @@ pgcopydb list extensions --source ${PGCOPYDB_SOURCE_PGURI}
 # that's the case is the existence of the pgcopydb.sentinel table on the
 # source database.
 #
-sql="select 1 from pg_class c join pg_namespace n on n.oid = c.relnamespace where relname = 'sentinel' and nspname = 'pgcopydb' union all select 0 limit 1"
-while [ `psql -At -d ${PGCOPYDB_SOURCE_PGURI} -c "${sql}"` -ne 1 ]
+while [ ! -s ${TMPDIR}/pgcopydb/schema/source.db ]
 do
     sleep 1
 done
